# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.

from sqlalchemy import Column, Integer, Text, String, Boolean, MetaData, Enum
from sqlalchemy.ext.declarative import declarative_base

from sickrage.core.databases import SRDatabase, SRDatabaseBase
from sickrage.core.enums import SeriesProviderID

class CacheDB(SRDatabase):
    base = declarative_base(cls=SRDatabaseBase)

    def __init__(self, db_type, db_prefix, db_host, db_port, db_username, db_password):
        super(CacheDB, self).__init__('cache', db_type, db_prefix, db_host, db_port, db_username, db_password)

    def initialize(self):
        self.base.metadata.create_all(self.engine)

    def cleanup(self):
        def remove_duplicates_from_last_search_table():
            found = []

            session = self.session()

            for x in session.query(CacheDB.LastSearch).all():
                if x.provider in found:
                    x.delete()
                    session.commit()
                else:
                    found.append(x.provider)

        def remove_duplicates_from_scene_name_table():
            found = []

            session = self.session()

            for x in session.query(CacheDB.SceneName).all():
                if (x.series_id, x.name) in found:
                    x.delete()
                    session.commit()
                else:
                    found.append((x.series_id, x.name))

        remove_duplicates_from_last_search_table()
        # remove_duplicates_from_scene_name_table()

    class LastUpdate(base):
        __tablename__ = 'last_update'

        provider = Column(String(32), primary_key=True)
        time = Column(Integer)

    class LastSearch(base):
        __tablename__ = 'last_search'

        provider = Column(String(32), primary_key=True)
        time = Column(Integer)

    class SceneName(base):
        __tablename__ = 'scene_names'

        id = Column(Integer, primary_key=True)
        series_id = Column(Integer)
        name = Column(Text)

    class NetworkTimezone(base):
        __tablename__ = 'network_timezones'

        network_name = Column(String(256), primary_key=True)
        timezone = Column(Text)

    class Provider(base):
        __tablename__ = 'providers'

        id = Column(Integer, primary_key=True)
        provider = Column(Text)
        name = Column(Text)
        season = Column(Integer)
        episodes = Column(Text)
        series_id = Column(Integer)
        series_provider_id = Column(Enum(SeriesProviderID))
        url = Column(String(256), index=True, unique=True)
        time = Column(Integer)
        quality = Column(Integer)
        release_group = Column(Text)
        version = Column(Integer, default=-1)
        seeders = Column(Integer)
        leechers = Column(Integer)
        size = Column(Integer)

    class OAuth2Token(base):
        __tablename__ = 'oauth2_token'

        id = Column(Integer, primary_key=True)
        access_token = Column(String(255), unique=True, nullable=False)
        refresh_token = Column(String(255), index=True)
        expires_in = Column(Integer, nullable=False, default=0)
        expires_at = Column(Integer, nullable=False, default=0)
        scope = Column(Text, default="")
        session_state = Column(Text, default="")
        token_type = Column(Text, default="bearer")

    class Announcements(base):
        __tablename__ = 'announcements'

        id = Column(Integer, primary_key=True)
        hash = Column(String(255), unique=True, nullable=False)
        seen = Column(Boolean, default=False)



# Copyright (c) 2013 Chris Lucas, <chris@chrisjlucas.com>
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# from rtorrent.rpc import Method
import rtorrentlib.rpc

from rtorrentlib.common import safe_repr

Method = rtorrentlib.rpc.Method


class Peer:
    """Represents an individual peer within a L{Torrent} instance."""
    def __init__(self, _rt_obj, info_hash, **kwargs):
        self._rt_obj = _rt_obj
        self.info_hash = info_hash  # : info hash for the torrent the peer is associated with
        for k in kwargs.keys():
            setattr(self, k, kwargs.get(k, None))

        self.rpc_id = "{0}:p{1}".format(
            self.info_hash, self.id)  # : unique id to pass to rTorrent

    def __repr__(self):
        return safe_repr("Peer(id={0})", self.id)

    def update(self):
        """Refresh peer data

        @note: All fields are stored as attributes to self.

        @return: None
        """
        multicall = rtorrentlib.rpc.Multicall(self)
        retriever_methods = [m for m in methods
                             if m.is_retriever() and m.is_available(self._rt_obj)]
        for method in retriever_methods:
            multicall.add(method, self.rpc_id)

        multicall.call()

methods = [
    # RETRIEVERS
    Method(Peer, 'is_preferred', 'p.is_preferred',
           boolean=True,
           ),
    Method(Peer, 'get_down_rate', 'p.down_rate'),
    Method(Peer, 'is_unwanted', 'p.is_unwanted',
           boolean=True,
           ),
    Method(Peer, 'get_peer_total', 'p.peer_total'),
    Method(Peer, 'get_peer_rate', 'p.peer_rate'),
    Method(Peer, 'get_port', 'p.port'),
    Method(Peer, 'is_snubbed', 'p.is_snubbed',
           boolean=True,
           ),
    Method(Peer, 'get_id_html', 'p.id_html'),
    Method(Peer, 'get_up_rate', 'p.up_rate'),
    Method(Peer, 'is_banned', 'p.banned',
           boolean=True,
           ),
    Method(Peer, 'get_completed_percent', 'p.completed_percent'),
    Method(Peer, 'get_id', 'p.id'),
    Method(Peer, 'is_obfuscated', 'p.is_obfuscated',
           boolean=True,
           ),
    Method(Peer, 'get_down_total', 'p.down_total'),
    Method(Peer, 'get_client_version', 'p.client_version'),
    Method(Peer, 'get_address', 'p.address'),
    Method(Peer, 'is_incoming', 'p.is_incoming',
           boolean=True,
           ),
    Method(Peer, 'is_encrypted', 'p.is_encrypted',
           boolean=True,
           ),
    Method(Peer, 'get_options_str', 'p.options_str'),
    Method(Peer, 'get_client_version', 'p.client_version'),
    Method(Peer, 'get_up_total', 'p.up_total'),

    # MODIFIERS
]



# coding=utf-8
# Author: Gonçalo M. (aka duramato/supergonkas) <supergonkas@gmail.com>
#
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE. If not, see <http://www.gnu.org/licenses/>.


import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import convert_size
from sickrage.search_providers import TorrentProvider


class HD4FreeProvider(TorrentProvider):
    def __init__(self):
        super(HD4FreeProvider, self).__init__('HD4Free', 'https://hd4free.xyz', True)

        self._urls.update({
            'search': '{base_url}/searchapi.php'.format(**self._urls)
        })

        # custom settings
        self.custom_settings = {
            'username': '',
            'api_key': '',
            'freeleech': False,
            'minseed': 0,
            'minleech': 0
        }

        self.cache = TVCache(self)

    def _check_auth(self):
        if self.custom_settings['username'] and self.custom_settings['api_key']:
            return True

        sickrage.app.log.warning('Your authentication credentials for {} are missing, check your config.'.format(self.name))

        return False

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []
        if not self._check_auth:
            return results

        search_params = {
            'tv': 'true',
            'username': self.custom_settings['username'],
            'apikey': self.custom_settings['api_key']
        }

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {0}".format(mode))
            for search_string in search_strings[mode]:
                if self.custom_settings['freeleech']:
                    search_params['fl'] = 'true'
                else:
                    search_params.pop('fl', '')

                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: {}".format(search_string.strip()))
                    search_params['search'] = search_string
                else:
                    search_params.pop('search', '')

                resp = self.session.get(self.urls['search'], params=search_params)
                if not resp or not resp.content:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                try:
                    data = resp.json()
                except ValueError:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(data, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        error = data.get('error')
        if error:
            sickrage.app.log.debug(error)
            return results

        try:
            if data['0']['total_results'] == 0:
                sickrage.app.log.debug("Provider has no results for this search")
                return results
        except Exception:
            return results

        for i in data:
            try:
                title = data[i]["release_name"]
                download_url = data[i]["download_url"]
                if not all([title, download_url]):
                    continue

                seeders = data[i]["seeders"]
                leechers = data[i]["leechers"]

                torrent_size = str(data[i]["size"]) + ' MB'
                size = convert_size(torrent_size, -1)

                results += [
                    {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                ]

                if mode != 'RSS':
                    sickrage.app.log.debug("Found result: {}".format(title))
            except Exception:
                sickrage.app.log.error("Failed parsing provider")

        return results



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

"""Initial migration

Revision ID: 20
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import datetime

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '20'
down_revision = '19'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    imdb_info = sa.Table('imdb_info', meta, autoload=True)

    with op.get_context().begin_transaction():
        for row in conn.execute(imdb_info.select()):
            conn.execute(f'UPDATE imdb_info SET last_update = "" WHERE imdb_info.indexer_id = {row.indexer_id}')

    with op.batch_alter_table("imdb_info") as batch_op:
        batch_op.alter_column('last_update', type_=sa.DateTime(timezone=True))

    with op.get_context().begin_transaction():
        for row in conn.execute(imdb_info.select()):
            conn.execute(f'UPDATE imdb_info SET last_update = {sa.func.current_timestamp()} WHERE imdb_info.indexer_id = {row.indexer_id}')

def downgrade():
    pass



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
from urllib.parse import parse_qsl

import oauth2
import twitter

import sickrage
from sickrage.notification_providers import NotificationProvider


class TwitterNotification(NotificationProvider):
    consumer_key = "vHHtcB6WzpWDG6KYlBMr8g"
    consumer_secret = "zMqq5CB3f8cWKiRO2KzWPTlBanYmV0VYxSXZ0Pxds0E"

    REQUEST_TOKEN_URL = 'https://api.twitter.com/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://api.twitter.com/oauth/access_token'
    AUTHORIZATION_URL = 'https://api.twitter.com/oauth/authorize'
    SIGNIN_URL = 'https://api.twitter.com/oauth/authenticate'

    def __init__(self):
        super(TwitterNotification, self).__init__()
        self.name = 'twitter'

    def notify_snatch(self, ep_name):
        if sickrage.app.config.twitter.notify_on_snatch:
            self._notifyTwitter(self.notifyStrings[self.NOTIFY_SNATCH] + ': ' + ep_name)

    def notify_download(self, ep_name):
        if sickrage.app.config.twitter.notify_on_download:
            self._notifyTwitter(self.notifyStrings[self.NOTIFY_DOWNLOAD] + ': ' + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.twitter.notify_on_subtitle_download:
            self._notifyTwitter(self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD] + ' ' + ep_name + ": " + lang)

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.twitter.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._notifyTwitter(title + " - " + update_text + new_version)

    def test_notify(self):
        return self._notifyTwitter("This is a test notification from SiCKRAGE", force=True)

    def _get_authorization(self):

        signature_method_hmac_sha1 = oauth2.SignatureMethod_HMAC_SHA1()
        oauth_consumer = oauth2.Consumer(key=self.consumer_key, secret=self.consumer_secret)
        oauth_client = oauth2.Client(oauth_consumer)

        sickrage.app.log.debug('Requesting temp token from Twitter')

        resp, content = oauth_client.request(self.REQUEST_TOKEN_URL, 'GET')

        if resp['status'] != '200':
            sickrage.app.log.error('Invalid response from Twitter requesting temp token: %s' % resp['status'])
        else:
            request_token = dict(parse_qsl(content))

            sickrage.app.config.twitter.username = request_token['oauth_token']
            sickrage.app.config.twitter.password = request_token['oauth_token_secret']

            return self.AUTHORIZATION_URL + "?oauth_token=" + request_token['oauth_token']

    def _get_credentials(self, key):
        request_token = {'oauth_token': sickrage.app.config.twitter.username,
                         'oauth_token_secret': sickrage.app.config.twitter.password,
                         'oauth_callback_confirmed': 'true'}

        token = oauth2.Token(request_token['oauth_token'], request_token['oauth_token_secret'])
        token.set_verifier(key)

        sickrage.app.log.debug('Generating and signing request for an access token using key ' + key)

        signature_method_hmac_sha1 = oauth2.SignatureMethod_HMAC_SHA1()
        oauth_consumer = oauth2.Consumer(key=self.consumer_key, secret=self.consumer_secret)
        sickrage.app.log.debug('oauth_consumer: ' + str(oauth_consumer))
        oauth_client = oauth2.Client(oauth_consumer, token)
        sickrage.app.log.debug('oauth_client: ' + str(oauth_client))
        resp, content = oauth_client.request(self.ACCESS_TOKEN_URL, method='POST', body='oauth_verifier=%s' % key)
        sickrage.app.log.debug('resp, content: ' + str(resp) + ',' + str(content))

        access_token = dict(parse_qsl(content))
        sickrage.app.log.debug('access_token: ' + str(access_token))

        sickrage.app.log.debug('resp[status] = ' + str(resp['status']))
        if resp['status'] != '200':
            sickrage.app.log.error('The request for a token with did not succeed: ' + str(resp['status']))
            return False
        else:
            sickrage.app.log.debug('Your Twitter Access Token key: %s' % access_token['oauth_token'])
            sickrage.app.log.debug('Access Token secret: %s' % access_token['oauth_token_secret'])
            sickrage.app.config.twitter.username = access_token['oauth_token']
            sickrage.app.config.twitter.password = access_token['oauth_token_secret']
            return True

    def _send_tweet(self, message=None):

        username = self.consumer_key
        password = self.consumer_secret
        access_token_key = sickrage.app.config.twitter.username
        access_token_secret = sickrage.app.config.twitter.password

        sickrage.app.log.debug("Sending tweet: {}".format(message))

        api = twitter.Api(username, password, access_token_key, access_token_secret)

        try:
            api.PostUpdate(message[:279])
        except Exception as e:
            sickrage.app.log.error("Error Sending Tweet: {}".format(e))
            return False

        return True

    def _send_dm(self, message=None):

        username = self.consumer_key
        password = self.consumer_secret
        dmdest = sickrage.app.config.twitter.dm_to
        access_token_key = sickrage.app.config.twitter.username
        access_token_secret = sickrage.app.config.twitter.password

        sickrage.app.log.debug("Sending DM: @{}: {}".format(dmdest, message))

        api = twitter.Api(username, password, access_token_key, access_token_secret)

        try:
            api.PostDirectMessage(message, screen_name=dmdest)
        except Exception as e:
            sickrage.app.log.error("Error Sending Tweet (DM): {}".format(e))
            return False

        return True

    def _notifyTwitter(self, message='', force=False):
        prefix = sickrage.app.config.twitter.prefix

        if not sickrage.app.config.twitter.enable and not force:
            return False

        if sickrage.app.config.twitter.use_dm and sickrage.app.config.twitter.dm_to:
            return self._send_dm(prefix + ": " + message)
        else:
            return self._send_tweet(prefix + ": " + message)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import sickrage
from sickrage.core.enums import ProcessMethod
from sickrage.core.webserver.handlers.api.v2 import ApiV2BaseHandler
from sickrage.core.webserver.handlers.api.v2.postprocess.schemas import PostProcessSchema


class Apiv2PostProcessHandler(ApiV2BaseHandler):
    def get(self):
        """Postprocess TV show video files"
        ---
        tags: [Post-Processing]
        summary: Manually post-process the files in the download folder
        description: Manually post-process the files in the download folder
        parameters:
        - in: query
          schema:
            PostProcessSchema
        responses:
          200:
            description: Success payload containing postprocess information
            content:
              application/json:
                schema:
                  PostProcessSuccessSchema
          400:
            description: Bad request; Check `errors` for any validation errors
            content:
              application/json:
                schema:
                  BadRequestSchema
          401:
            description: Returned if your JWT token is missing or expired
            content:
              application/json:
                schema:
                  NotAuthorizedSchema
        """
        path = self.get_argument("path", sickrage.app.config.general.tv_download_dir)
        nzb_name = self.get_argument("nzbName", None)
        process_method = self.get_argument("processMethod", ProcessMethod.COPY.name)
        proc_type = self.get_argument("type", 'manual')
        delete = self._parse_boolean(self.get_argument("delete", 'false'))
        failed = self._parse_boolean(self.get_argument("failed", 'false'))
        is_priority = self._parse_boolean(self.get_argument("isPriority", 'false'))
        return_data = self._parse_boolean(self.get_argument("returnData", 'false'))
        force_replace = self._parse_boolean(self.get_argument("forceReplace", 'false'))
        force_next = self._parse_boolean(self.get_argument("forceNext", 'false'))

        validation_errors = self._validate_schema(PostProcessSchema, self.request.arguments)
        if validation_errors:
            return self._bad_request(error=validation_errors)

        if not path and not sickrage.app.config.general.tv_download_dir:
            return self._bad_request(error={"path": "You need to provide a path or set TV Download Dir"})

        json_data = sickrage.app.postprocessor_queue.put(path, nzbName=nzb_name, process_method=ProcessMethod[process_method.upper()], force=force_replace,
                                                         is_priority=is_priority, delete_on=delete, failed=failed, proc_type=proc_type, force_next=force_next)

        if 'Processing succeeded' not in json_data and 'Successfully processed' not in json_data:
            return self._bad_request(error=json_data)

        return self.json_response({'data': json_data if return_data else ''})



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import os
import pickle
import random
import sqlite3
import threading
from time import sleep

import alembic.command
import alembic.config
import alembic.script
import sqlalchemy
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory
from attrdict import AttrDict
from sqlalchemy import create_engine, event, inspect, MetaData, Index, TypeDecorator
from sqlalchemy.engine import Engine, reflection
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.serializer import loads, dumps
from sqlalchemy.orm import sessionmaker, mapper, scoped_session
from sqlalchemy.sql.ddl import CreateTable, CreateIndex
from sqlalchemy.util import KeyedTuple

import sickrage


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    if not isinstance(dbapi_connection, sqlite3.Connection):
        return

    old_isolation = dbapi_connection.isolation_level
    dbapi_connection.isolation_level = None
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA journal_mode=WAL")
    cursor.close()
    dbapi_connection.isolation_level = old_isolation


@event.listens_for(mapper, "init")
def instant_defaults_listener(target, args, kwargs):
    for key, column in inspect(target.__class__).columns.items():
        if column.default is not None:
            if callable(column.default.arg):
                setattr(target, key, column.default.arg(target))
            else:
                setattr(target, key, column.default.arg)


class IntFlag(TypeDecorator):
    impl = sqlalchemy.types.Integer()

    def __init__(self, enum):
        self.enum = enum

    def process_bind_param(self, value, dialect):
        return int(value) if value is not None else None

    def process_result_value(self, value, dialect):
        return self.enum(value) if value is not None else None


class ContextSession(sqlalchemy.orm.Session):
    """:class:`sqlalchemy.orm.Session` which can be used as context manager"""

    def __init__(self, *args, **kwargs):
        super(ContextSession, self).__init__(*args, **kwargs)
        self._lock = threading.RLock()
        self.max_attempts = 50

    def commit(self, close=False):
        statement = None
        params = None
        for i in range(self.max_attempts):
            try:
                if statement and params:
                    self.bind.execute(statement, params)
                super(ContextSession, self).commit()
            except OperationalError as e:
                self.rollback()

                if 'database is locked' not in str(e):
                    raise

                statement = e.statement
                params = e.params

                timer = random.randint(10, 30)
                sickrage.app.log.debug('Retrying database commit in {}s, attempt {}'.format(timer, i))
                sleep(timer)
            except Exception as e:
                self.rollback()
                raise
            else:
                break
            finally:
                if close:
                    self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class SRDatabaseBase(object):
    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def as_attrdict(self):
        return AttrDict(self.as_dict())

    def update(self, **kwargs):
        primary_keys = [pk.name for pk in self.__table__.primary_key]
        for key, value in kwargs.items():
            if key not in primary_keys:
                setattr(self, key, value)


class SRDatabase(object):
    def __init__(self, name, db_type='sqlite', db_prefix='sickrage', db_host='localhost', db_port='3306', db_username='sickrage', db_password='sickrage'):
        self.name = name
        self.db_type = db_type
        self.db_prefix = db_prefix
        self.db_host = db_host
        self.db_port = db_port
        self.db_username = db_username
        self.db_password = db_password

        self.db_path = os.path.join(sickrage.app.data_dir, '{}.db'.format(self.name))
        self.db_migrations_path = os.path.join(os.path.dirname(__file__), self.name, 'migrations')

        self.session = scoped_session(sessionmaker(class_=ContextSession, bind=self.engine))

    @property
    def engine(self):
        if self.db_type == 'sqlite':
            return create_engine('sqlite:///{}'.format(self.db_path), echo=False, connect_args={'check_same_thread': False, 'timeout': 30})
        elif self.db_type == 'mysql':
            mysql_engine = create_engine('mysql+pymysql://{}:{}@{}:{}/'.format(self.db_username, self.db_password, self.db_host, self.db_port), echo=False)
            mysql_engine.execute(f"CREATE DATABASE IF NOT EXISTS {self.db_prefix}_{self.name}")
            return create_engine(
                'mysql+pymysql://{}:{}@{}:{}/{}_{}'.format(self.db_username, self.db_password, self.db_host, self.db_port, self.db_prefix, self.name),
                echo=False)

    @property
    def version(self):
        context = MigrationContext.configure(self.engine)
        current_rev = context.get_current_revision()
        return current_rev

    def setup(self):
        if self.engine.dialect.has_table(self.engine, 'migrate_version'):
            migrate_version = self.engine.execute("select version from migrate_version").fetchone().version
            alembic.command.stamp(self.get_alembic_config(), str(migrate_version))
            self.engine.execute("drop table migrate_version")

        if not self.engine.dialect.has_table(self.engine, 'alembic_version'):
            alembic.command.stamp(self.get_alembic_config(), 'head')
            sickrage.app.log.info("Performing initialization on {} database".format(self.name))
            self.initialize()

        # perform integrity check
        sickrage.app.log.info("Performing integrity check on {} database".format(self.name))
        self.integrity_check()

        # upgrade database
        sickrage.app.log.info("Performing upgrades on {} database".format(self.name))
        self.upgrade()

        # cleanup
        sickrage.app.log.info("Performing cleanup on {} database".format(self.name))
        self.cleanup()

        # free up space
        sickrage.app.log.info("Performing vacuum on {} database".format(self.name))
        self.vacuum()

    def initialize(self):
        pass

    def upgrade(self):
        db_version = int(self.version)
        alembic_version = int(ScriptDirectory.from_config(self.get_alembic_config()).get_current_head())

        backup_filename = os.path.join(sickrage.app.data_dir, f'{self.name}_db_backup_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}.json')

        if db_version < alembic_version:
            # temp code to resolve a migration bug introduced from v10.0.0, fixed in v10.0.2+
            if db_version < 21 and self.name == 'main':
                if self.engine.dialect.has_table(self.engine, 'indexer_mapping') and self.engine.dialect.has_table(self.engine, 'series_provider_mapping'):
                    sickrage.app.log.debug('Found offending series_provider_mapping table, removing!')
                    metadata = MetaData(self.engine, reflect=True)
                    table = metadata.tables.get('series_provider_mapping')
                    table.drop(self.engine)

            sickrage.app.log.info(f'Backing up {self.name} database v{db_version}')
            self.backup(backup_filename)

            sickrage.app.log.info(f'Upgrading {self.name} database to v{alembic_version}')
            alembic.command.upgrade(self.get_alembic_config(), 'head')

    def get_alembic_config(self):
        config = alembic.config.Config()
        config.set_main_option('script_location', self.db_migrations_path)
        config.set_main_option('sqlalchemy.url', str(self.engine.url))
        config.set_main_option('url', str(self.engine.url))
        return config

    def get_metadata(self):
        return MetaData(bind=self.engine, reflect=True)

    def get_base(self):
        base = automap_base(metadata=self.get_metadata())
        base.prepare()
        return base

    def integrity_check(self):
        if self.db_type == 'sqlite':
            if self.session().scalar("PRAGMA integrity_check") != "ok":
                sickrage.app.log.fatal(
                    f"{self.name.capitalize()} database file {self.db_path} is damaged, please restore a backup or delete the database file and restart SiCKRAGE")

    def cleanup(self):
        pass

    def vacuum(self):
        self.engine.execute("VACUUM")

    def backup(self, filename):
        meta = self.get_metadata()

        backup_dict = {
            'schema': {},
            'indexes': {},
            'data': {},
            'version': self.version
        }

        for table_name, table_object in meta.tables.items():
            sickrage.app.log.info(f'Backing up {self.name} database table {table_name} schema')
            backup_dict['indexes'].update({table_name: []})
            backup_dict['schema'].update({table_name: str(CreateTable(table_object))})
            backup_dict['data'].update({table_name: dumps(self.session().query(table_object).all(), protocol=pickle.DEFAULT_PROTOCOL)})
            for index in reflection.Inspector.from_engine(self.engine).get_indexes(table_name):
                cols = [table_object.c[col] for col in index['column_names']]
                idx = Index(index['name'], *cols)
                backup_dict['indexes'][table_name].append(str(CreateIndex(idx)))

        with open(filename, 'wb') as fh:
            pickle.dump(backup_dict, fh, protocol=pickle.DEFAULT_PROTOCOL)

    def restore(self, filename):
        session = self.session()

        with open(filename, 'rb') as fh:
            backup_dict = pickle.load(fh)

            # drop all tables
            self.get_base().metadata.drop_all()

            # restore schema
            if backup_dict.get('schema', None):
                for table_name, schema in backup_dict['schema'].items():
                    sickrage.app.log.info(f'Restoring {self.name} database table {table_name} schema')
                    session.execute(schema)
                session.commit()

            # restore indexes
            if backup_dict.get('indexes', None):
                for table_name, indexes in backup_dict['indexes'].items():
                    sickrage.app.log.info(f'Restoring {self.name} database table {table_name} indexes')
                    for index in indexes:
                        session.execute(index)
                session.commit()

            # restore data
            if backup_dict.get('data', None):
                base = self.get_base()
                meta = self.get_metadata()
                for table_name, data in backup_dict['data'].items():
                    sickrage.app.log.info(f'Restoring {self.name} database table {table_name} data')
                    table = base.classes[table_name]
                    session.query(table).delete()

                    rows = []
                    for row in loads(data, meta, session):
                        if isinstance(row, KeyedTuple):
                            rows.append(row._asdict())
                    session.bulk_insert_mappings(table, rows)
                session.commit()



#!/usr/bin/env python
# coding=utf-8
#
# This file is part of aDBa.
#
# aDBa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aDBa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with aDBa.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
import string

from six import string_types

from . import aniDBfileInfo as fileInfo
from .aniDBerrors import *
from .aniDBfileInfo import read_anidb_xml
from .aniDBmaper import AniDBMaper
from .aniDBtvDBmaper import TvDBMap

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class aniDBabstractObject(object):
    def __init__(self, aniDB, load=False):
        self.laoded = False
        self.set_connection(aniDB)
        if load:
            self.load_data()

    def set_connection(self, aniDB):
        self.aniDB = aniDB

    def _fill(self, dataline):
        for key in dataline:
            try:
                tmp_list = dataline[key].split("'")
                if len(tmp_list) > 1:
                    new_list = []
                    for i in tmp_list:
                        try:
                            new_list.append(int(i))
                        except:
                            new_list.append(i)
                    self.__dict__[key] = new_list
                    continue
            except:
                pass
            try:
                self.__dict__[key] = int(dataline[key])
            except:
                # self.__dict__[key] = text_type(dataline[key], "utf-8")
                self.__dict__[key] = dataline[key]
            key = property(lambda x: dataline[key])

    def __getattr__(self, name):
        try:
            return object.__getattribute__(self, name)
        except:
            return None

    def _build_names(self):
        names = []
        names = self._easy_extend(names, self.english_name)
        names = self._easy_extend(names, self.short_name_list)
        names = self._easy_extend(names, self.synonym_list)
        names = self._easy_extend(names, self.other_name)

        self.allNames = names

    @staticmethod
    def _easy_extend(initialList, item):
        if item:
            if isinstance(item, list):
                initialList.extend(item)
            elif isinstance(item, string_types):
                initialList.append(item)

        return initialList

    def load_data(self):
        return False

    def add_notification(self):
        """
        type - Type of notification: type=>  0=all, 1=new, 2=group, 3=complete
        priority - low = 0, medium = 1, high = 2 (unconfirmed)

        """
        if self.aid:
            self.aniDB.notifyadd(aid=self.aid, type=1, priority=1)


class Anime(aniDBabstractObject):
    def __init__(self, aniDB, name=None, aid=None, tvdbid=None, paramsA=None, autoCorrectName=False, load=False, cache_path=None):

        self.maper = AniDBMaper()

        if cache_path and not os.path.exists(cache_path):
            os.makedirs(cache_path)

        self.tvDBMap = TvDBMap(cache_path)
        self.allAnimeXML = None

        self.name = name
        self.aid = aid
        self.tvdb_id = tvdbid
        self.cache_path = cache_path

        if self.tvdb_id and not self.aid:
            self.aid = self.tvDBMap.get_anidb_for_tvdb(self.tvdb_id)

        if not (self.name or self.aid):
            raise AniDBIncorrectParameterError("No aid or name available")

        if not self.aid:
            self.aid = self._get_aid_from_xml(self.name)
        # if not self.name or autoCorrectName:
        #    self.name = self._get_name_from_xml(self.aid)

        if not (self.name or self.aid):
            raise ValueError

        if not self.tvdb_id:
            self.tvdb_id = self.tvDBMap.get_tvdb_for_anidb(self.aid)

        if not paramsA:
            self.bitCode = "b2f0e0fc000000"
            self.params = self.maper.getAnimeCodesA(self.bitCode)
        else:
            self.paramsA = paramsA
            self.bitCode = self.maper.getAnimeBitsA(self.paramsA)

        super(Anime, self).__init__(aniDB, load)

    def load_data(self):
        """load the data from anidb"""

        if not (self.name or self.aid):
            raise ValueError

        self.rawData = self.aniDB.anime(aid=self.aid, aname=self.name, amask=self.bitCode)
        if self.rawData.datalines:
            self._fill(self.rawData.datalines[0])
            self._builPreSequal()
            self.laoded = True

    def return_raw_data(self):
        """Returns all raw data from request"""
        return self.rawData

    def get_groups(self):
        if not self.aid:
            return []
        self.rawData = self.aniDB.groupstatus(aid=self.aid)
        self.release_groups = []
        for line in self.rawData.datalines:
            self.release_groups.append({"name": line["name"].decode("utf-8"),
                                        "rating": line["rating"],
                                        "range": line["episode_range"]
                                        })
        return self.release_groups

    # TODO: refactor and use the new functions in anidbFileinfo
    def _get_aid_from_xml(self, name):
        if not self.allAnimeXML:
            self.allAnimeXML = read_anidb_xml(self.cache_path)

        regex = re.compile('( \(\d{4}\))|[%s]' % re.escape(string.punctuation))  # remove any punctuation and e.g. ' (2011)'
        # regex = re.compile('[%s]'  % re.escape(string.punctuation)) # remove any punctuation and e.g. ' (2011)'
        name = regex.sub('', name.lower())
        last_aid = 0
        for element in self.allAnimeXML.iter():
            if element.get("aid", False):
                last_aid = int(element.get("aid"))
            if element.text:
                testname = regex.sub('', element.text.lower())

                if testname == name:
                    return last_aid
        return 0

    # TODO: refactor and use the new functions in anidbFileinfo
    def _get_name_from_xml(self, aid, onlyMain=True):
        if not self.allAnimeXML:
            self.allAnimeXML = read_anidb_xml(self.cache_path)

        for anime in self.allAnimeXML.findall("anime"):
            if int(anime.get("aid", False)) == aid:
                for title in anime.getiterator():
                    current_lang = title.get("{http://www.w3.org/XML/1998/namespace}lang", False)
                    current_type = title.get("type", False)
                    if (current_lang == "en" and not onlyMain) or current_type == "main":
                        return title.text
        return ""

    def _builPreSequal(self):
        if self.related_aid_list and self.related_aid_type:
            try:
                for i in range(len(self.related_aid_list)):
                    if self.related_aid_type[i] == 2:
                        self.__dict__["prequal"] = self.related_aid_list[i]
                    elif self.related_aid_type[i] == 1:
                        self.__dict__["sequal"] = self.related_aid_list[i]
            except:
                if self.related_aid_type == 2:
                    self.__dict__["prequal"] = self.related_aid_list
                elif self.str_related_aid_type == 1:
                    self.__dict__["sequal"] = self.related_aid_list


class Episode(aniDBabstractObject):
    def __init__(self, aniDB, number=None, epid=None, filePath=None, fid=None, epno=None, paramsA=None, paramsF=None, load=False, calculate=False):
        if not aniDB and not number and not epid and not filePath and not fid:
            return None

        self.maper = AniDBMaper()
        self.epid = epid
        self.filePath = filePath
        self.fid = fid
        self.epno = epno
        if calculate:
            (self.ed2k, self.size) = self._calculate_file_stuff(self.filePath)

        if not paramsA:
            self.bitCodeA = "C000F0C0"
            self.paramsA = self.maper.getFileCodesA(self.bitCodeA)
        else:
            self.paramsA = paramsA
            self.bitCodeA = self.maper.getFileBitsA(self.paramsA)

        if not paramsF:
            self.bitCodeF = "7FF8FEF8"
            self.paramsF = self.maper.getFileCodesF(self.bitCodeF)
        else:
            self.paramsF = paramsF
            self.bitCodeF = self.maper.getFileBitsF(self.paramsF)

        super(Episode, self).__init__(aniDB, load)

    def load_data(self):
        """load the data from anidb"""
        if self.filePath and not (self.ed2k or self.size):
            (self.ed2k, self.size) = self._calculate_file_stuff(self.filePath)

        self.rawData = self.aniDB.file(fid=self.fid, size=self.size, ed2k=self.ed2k, aid=self.aid, aname=None, gid=None, gname=None, epno=self.epno, fmask=self.bitCodeF, amask=self.bitCodeA)
        self._fill(self.rawData.datalines[0])
        self._build_names()
        self.laoded = True

    def add_to_mylist(self, state=None, viewed=None, source=None, storage=None, other=None):
        """
        state    - the location of the file
                viewed    - whether you have watched the file (0=unwatched,1=watched)
                source    - where you got the file (bittorrent,dc++,ed2k,...)
                storage    - for example the title of the cd you have this on
                other    - other data regarding this file

        structure of state:
                value    meaning
                0    unknown    - state is unknown or the user doesn't want to provide this information
                1    on hdd    - the file is stored on hdd
                2    on cd    - the file is stored on cd
                3    deleted    - the file has been deleted or is not available for other reasons (i.e. reencoded)
        """
        if self.filePath and not (self.ed2k or self.size):
            (self.ed2k, self.size) = self._calculate_file_stuff(self.filePath)
        try:
            self.aniDB.mylistadd(size=self.size, ed2k=self.ed2k, state=state, viewed=viewed, source=source, storage=storage, other=other)
        except Exception as e:
            logger.exception("Exception: %s", e)
        else:
            # TODO: add the name or something
            logger.info("Added the episode to anidb")

    def edit_to_mylist(self, state=None, viewed=None, source=None, storage=None, other=None):
        """
        state    - the location of the file
                viewed    - whether you have watched the file (0=unwatched,1=watched)
                source    - where you got the file (bittorrent,dc++,ed2k,...)
                storage    - for example the title of the cd you have this on
                other    - other data regarding this file

        structure of state:
                value    meaning
                0    unknown    - state is unknown or the user doesn't want to provide this information
                1    on hdd    - the file is stored on hdd
                2    on cd    - the file is stored on cd
                3    deleted    - the file has been deleted or is not available for other reasons (i.e. reencoded)
        """
        if self.filePath and not (self.ed2k or self.size):
            (self.ed2k, self.size) = self._calculate_file_stuff(self.filePath)
        try:
            edit_response = self.aniDB.mylistadd(size=self.size, ed2k=self.ed2k, edit=1, state=state, viewed=viewed, source=source, storage=storage, other=other)
        except Exception as e:
            logger.exception("Exception: %s", e)
        # handling the case that the entry is not in anidb yet, non ideal to check the string but isinstance is having issue
        # currently raises an exception for less changes in the code, unsure if this is the ideal way to do so
        if edit_response.codestr == "NO_SUCH_MYLIST_ENTRY":
            logger.info("attempted an edit before add")
            raise AniDBError("Attempted to edit file without adding")
        else:
            logger.info("Edited the episode in anidb")

    def delete_from_mylist(self):
        if self.filePath and not (self.ed2k or self.size):
            (self.ed2k, self.size) = self._calculate_file_stuff(self.filePath)
        try:
            self.aniDB.mylistdel(size=self.size, ed2k=self.ed2k)
        except Exception as e:
            logger.exception("Exception: %s", e)
        else:
            logger.info("Deleted the episode from anidb")

    @staticmethod
    def _calculate_file_stuff(filePath):
        if not filePath:
            return None, None
        logger.info("Calculating the ed2k. Please wait...")
        ed2k = fileInfo.get_ED2K(filePath)
        size = fileInfo.get_file_size(filePath)
        return ed2k, size



import functools
import threading

import sickrage


class RSSCacheUpdater(object):
    def __init__(self):
        super(RSSCacheUpdater, self).__init__()
        self.name = "RSSCACHE-UPDATER"
        self.lock = threading.Lock()
        self.running = False

    def task(self, force=False):
        if self.running or not sickrage.app.config.general.enable_rss_cache and not force:
            return

        try:
            self.running = True

            for providerID, providerObj in sickrage.app.search_providers.sort().items():
                if providerObj.is_enabled:
                    threading.currentThread().setName('{}::{}'.format(self.name, providerObj.name.upper()))
                    providerObj.cache.update(force)
        finally:
            self.running = False



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


from tornado.web import authenticated

import sickrage
from sickrage.core.enums import ProcessMethod
from sickrage.core.helpers import arg_to_bool
from sickrage.core.webserver.handlers.base import BaseHandler


class HomePostProcessHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('home/postprocess.mako',
                           title=_('Post Processing'),
                           header=_('Post Processing'),
                           topmenu='home',
                           controller='home',
                           action='postprocess')


class HomeProcessEpisodeHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return "Please use our API instead for post-processing"

    @authenticated
    def post(self, *args, **kwargs):
        pp_options = {
            'proc_dir': self.get_argument('proc_dir'),
            'nzbname': self.get_argument('nzbname', ''),
            'process_method': self.get_argument('process_method', sickrage.app.config.general.process_method),
            'proc_type': self.get_argument('proc_type', 'manual'),
            'force': arg_to_bool(self.get_argument('force', 'false')),
            'is_priority': arg_to_bool(self.get_argument('is_priority', 'false')),
            'delete_on': arg_to_bool(self.get_argument('delete_on', 'false')),
            'force_next': arg_to_bool(self.get_argument('force_next', 'false')),
            'failed': arg_to_bool(self.get_argument('failed', 'false')),
            'quiet': arg_to_bool(self.get_argument('quiet', 'false')),
        }

        proc_dir = pp_options.pop("proc_dir")
        quiet = pp_options.pop("quiet")

        if not proc_dir:
            return self.redirect("/home/postprocess/")

        if not isinstance(pp_options['process_method'], ProcessMethod):
            pp_options['process_method'] = ProcessMethod[pp_options['process_method'].upper()]

        result = sickrage.app.postprocessor_queue.put(proc_dir, **pp_options)
        if quiet:
            return result

        return self._genericMessage(_("Postprocessing results"), result.replace("\n", "<br>\n"))



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.

import sickrage
from sickrage.core.websession import WebSession
from sickrage.notification_providers import NotificationProvider


class TelegramNotification(NotificationProvider):
    """
    A notifier for Telegram
    """

    def __init__(self):
        super(TelegramNotification, self).__init__()
        self.name = 'telegram'

    def test_notify(self, id=None, api_key=None):
        """
        Send a test notification
        :param id: The Telegram user/group id to send the message to
        :param api_key: Your Telegram bot API token
        :returns: the notification
        """
        return self._notify_telegram('Test', 'This is a test notification from SickRage', id, api_key, force=True)

    def _send_telegram_msg(self, title, msg, id=None, api_key=None):
        """
        Sends a Telegram notification

        :param title: The title of the notification to send
        :param msg: The message string to send
        :param id: The Telegram user/group id to send the message to
        :param api_key: Your Telegram bot API token

        :returns: True if the message succeeded, False otherwise
        """
        id = sickrage.app.config.telegram.user_id or id
        api_key = sickrage.app.config.telegram.apikey or api_key

        payload = {'chat_id': id, 'text': '{} : {}'.format(title, msg)}

        telegram_api = 'https://api.telegram.org/bot{}/{}'

        try:
            resp = WebSession().post(telegram_api.format(api_key, 'sendMessage'), json=payload).json()
            success = resp['ok']
            message = 'Telegram message sent successfully.' if success else '{} {}'.format(resp['error_code'], resp['description'])
        except Exception as e:
            success = False
            message = 'Error while sending Telegram message: {} '.format(e)

        sickrage.app.log.info(message)
        return success, message

    def notify_snatch(self, ep_name, title=None):
        """
        Sends a Telegram notification when an episode is snatched

        :param ep_name: The name of the episode snatched
        :param title: The title of the notification to send
        """
        if not title:
            title = self.notifyStrings[self.NOTIFY_SNATCH]

        if sickrage.app.config.telegram.notify_on_snatch:
            self._notify_telegram(title, ep_name)

    def notify_download(self, ep_name, title=None):
        """
        Sends a Telegram notification when an episode is downloaded

        :param ep_name: The name of the episode downloaded
        :param title: The title of the notification to send
        """
        if not title:
            title = self.notifyStrings[self.NOTIFY_DOWNLOAD]

        if sickrage.app.config.telegram.notify_on_download:
            self._notify_telegram(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=None):
        """
        Sends a Telegram notification when subtitles for an episode are downloaded

        :param ep_name: The name of the episode subtitles were downloaded for
        :param lang: The language of the downloaded subtitles
        :param title: The title of the notification to send
        """
        if not title:
            title = self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD]

        if sickrage.app.config.telegram.notify_on_subtitle_download:
            self._notify_telegram(title, '{}: {}'.format(ep_name, lang))

    def notify_version_update(self, new_version='??'):
        """
        Sends a Telegram notification for git updates

        :param new_version: The new version available from git
        """
        if sickrage.app.config.telegram.enable:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._notify_telegram(title, update_text + new_version)

    def notify_login(self, ipaddress=''):
        """
        Sends a Telegram notification on login

        :param ipaddress: The ip address the login is originating from
        """
        if sickrage.app.config.telegram.enable:
            update_text = self.notifyStrings[self.NOTIFY_LOGIN_TEXT]
            title = self.notifyStrings[self.NOTIFY_LOGIN]
            self._notify_telegram(title, update_text.format(ipaddress))

    def _notify_telegram(self, title, message, id=None, api_key=None, force=False):
        """
        Sends a Telegram notification

        :param title: The title of the notification to send
        :param message: The message string to send
        :param id: The Telegram user/group id to send the message to
        :param api_key: Your Telegram bot API token
        :param force: Enforce sending, for instance for testing

        :returns: the message to send
        """

        if not (force or sickrage.app.config.telegram.enable):
            sickrage.app.log.debug('Notification for Telegram not enabled, skipping this notification')
            return False, 'Disabled'

        sickrage.app.log.debug('Sending a Telegram message for {}'.format(message))

        return self._send_telegram_msg(title, message, id, api_key)



"""Initial migration

Revision ID: 8
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import datetime

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '8'
down_revision = '7'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    failed_snatch_history = sa.Table('failed_snatch_history', meta, autoload=True)

    op.alter_column('failed_snatch_history', 'date', type_=sa.String(32))

    date_format = '%Y%m%d%H%M%S'

    with op.get_context().begin_transaction():
        for row in conn.execute(failed_snatch_history.select()):
            date = datetime.datetime.strptime(str(row.date), date_format)
            conn.execute(f'UPDATE failed_snatch_history SET date = {date} WHERE failed_snatch_history.id = {row.id}')

    op.alter_column('failed_snatch_history', 'date', type_=sa.DateTime)


def downgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    failed_snatch_history = sa.Table('failed_snatch_history', meta, autoload=True)

    op.alter_column('failed_snatch_history', 'date', type_=sa.String(32))

    with op.get_context().begin_transaction():
        for row in conn.execute(failed_snatch_history.select()):
            date = str(row.date.toordinal())
            conn.execute(f'UPDATE failed_snatch_history SET date = {date} WHERE failed_snatch_history.id = {row.id}')

    op.alter_column('failed_snatch_history', 'date', type_=sa.Integer)



"""Initial migration

Revision ID: 3
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '3'
down_revision = '2'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    tv_shows = sa.Table('tv_shows', meta, autoload=True)

    if not hasattr(tv_shows.c, 'last_proper_search'):
        op.add_column('tv_shows', sa.Column('last_proper_search', sa.Integer, default=0))


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import datetime
import glob
import os
import re
import shutil
import stat
import threading
import traceback

import send2trash
import sqlalchemy
from adba.aniDBAbstracter import Anime
from sqlalchemy import orm
from unidecode import unidecode

import sickrage
from sickrage.core.blackandwhitelist import BlackAndWhiteList
from sickrage.core.caches.image_cache import ImageCache
from sickrage.core.common import Quality, Qualities, EpisodeStatus
from sickrage.core.databases.main import MainDB
from sickrage.core.databases.main.schemas import TVShowSchema, IMDbInfoSchema, BlacklistSchema, WhitelistSchema
from sickrage.core.enums import SeriesProviderID
from sickrage.core.exceptions import ShowNotFoundException, EpisodeNotFoundException, EpisodeDeletedException, MultipleEpisodesInDatabaseException
from sickrage.core.helpers import list_media_files, is_media_file, try_int, safe_getattr, flatten
from sickrage.core.media.util import series_image, SeriesImageType
from sickrage.core.tv.episode import TVEpisode
from sickrage.series_providers.exceptions import SeriesProviderAttributeNotFound, SeriesProviderException


class TVShowID(object):
    def __init__(self, series_id, series_provider_id):
        self.series_id = series_id
        self.series_provider_id = series_provider_id

    @property
    def slug(self):
        return f'{self.series_id}-{self.series_provider_id.slug}'


class TVShow(object):
    def __init__(self, series_id, series_provider_id, lang='en', location=''):
        self.lock = threading.Lock()
        self._episodes = {}

        with sickrage.app.main_db.session() as session:
            try:
                query = session.query(MainDB.TVShow).filter_by(series_id=series_id, series_provider_id=series_provider_id).one()
                self._data_local = query.as_dict()
            except orm.exc.NoResultFound:
                self._data_local = MainDB.TVShow().as_dict()
                self._data_local.update(**{
                    'series_id': series_id,
                    'series_provider_id': series_provider_id,
                    'lang': lang,
                    'location': location
                })

                self.load_from_series_provider()

        sickrage.app.shows.update({(self.series_id, self.series_provider_id): self})

    @property
    def slug(self):
        return f'{self.series_id}-{self.series_provider_id.slug}'

    @property
    def series_id(self):
        return self._data_local['series_id']

    @series_id.setter
    def series_id(self, value):
        self._data_local['series_id'] = value

    @property
    def series_provider_id(self):
        return self._data_local['series_provider_id']

    @series_provider_id.setter
    def series_provider_id(self, value):
        self._data_local['series_provider_id'] = value

    @property
    def tvdb_id(self):
        if self.series_provider_id == SeriesProviderID.THETVDB:
            return self.series_id

    @property
    def name(self):
        return self._data_local['name']

    @name.setter
    def name(self, value):
        self._data_local['name'] = value

    @property
    def location(self):
        return self._data_local['location']

    @location.setter
    def location(self, value):
        self._data_local['location'] = value

    @property
    def network(self):
        return self._data_local['network']

    @network.setter
    def network(self, value):
        self._data_local['network'] = value

    @property
    def genre(self):
        return self._data_local['genre']

    @genre.setter
    def genre(self, value):
        self._data_local['genre'] = value

    @property
    def overview(self):
        return self._data_local['overview']

    @overview.setter
    def overview(self, value):
        self._data_local['overview'] = value

    @property
    def classification(self):
        return self._data_local['classification']

    @classification.setter
    def classification(self, value):
        self._data_local['classification'] = value

    @property
    def runtime(self):
        return self._data_local['runtime']

    @runtime.setter
    def runtime(self, value):
        self._data_local['runtime'] = value

    @property
    def quality(self):
        return self._data_local['quality']

    @quality.setter
    def quality(self, value):
        self._data_local['quality'] = value

    @property
    def airs(self):
        return self._data_local['airs']

    @airs.setter
    def airs(self, value):
        self._data_local['airs'] = value

    @property
    def status(self):
        return self._data_local['status']

    @status.setter
    def status(self, value):
        self._data_local['status'] = value

    @property
    def flatten_folders(self):
        return self._data_local['flatten_folders']

    @flatten_folders.setter
    def flatten_folders(self, value):
        self._data_local['flatten_folders'] = value

    @property
    def paused(self):
        return self._data_local['paused']

    @paused.setter
    def paused(self, value):
        self._data_local['paused'] = value

    @property
    def scene(self):
        return self._data_local['scene']

    @scene.setter
    def scene(self, value):
        self._data_local['scene'] = value

    @property
    def anime(self):
        return self._data_local['anime']

    @anime.setter
    def anime(self, value):
        self._data_local['anime'] = value

    @property
    def search_format(self):
        return self._data_local['search_format']

    @search_format.setter
    def search_format(self, value):
        self._data_local['search_format'] = value

    @property
    def subtitles(self):
        return self._data_local['subtitles']

    @subtitles.setter
    def subtitles(self, value):
        self._data_local['subtitles'] = value

    @property
    def dvd_order(self):
        return self._data_local['dvd_order']

    @dvd_order.setter
    def dvd_order(self, value):
        self._data_local['dvd_order'] = value

    @property
    def skip_downloaded(self):
        return self._data_local['skip_downloaded']

    @skip_downloaded.setter
    def skip_downloaded(self, value):
        self._data_local['skip_downloaded'] = value

    @property
    def startyear(self):
        return self._data_local['startyear']

    @startyear.setter
    def startyear(self, value):
        self._data_local['startyear'] = value

    @property
    def lang(self):
        return self._data_local['lang']

    @lang.setter
    def lang(self, value):
        self._data_local['lang'] = value

    @property
    def imdb_id(self):
        return self._data_local['imdb_id']

    @imdb_id.setter
    def imdb_id(self, value):
        self._data_local['imdb_id'] = value

    @property
    def rls_ignore_words(self):
        return self._data_local['rls_ignore_words']

    @rls_ignore_words.setter
    def rls_ignore_words(self, value):
        self._data_local['rls_ignore_words'] = value

    @property
    def rls_require_words(self):
        return self._data_local['rls_require_words']

    @rls_require_words.setter
    def rls_require_words(self, value):
        self._data_local['rls_require_words'] = value

    @property
    def default_ep_status(self):
        return self._data_local['default_ep_status']

    @default_ep_status.setter
    def default_ep_status(self, value):
        self._data_local['default_ep_status'] = value

    @property
    def sub_use_sr_metadata(self):
        return self._data_local['sub_use_sr_metadata']

    @sub_use_sr_metadata.setter
    def sub_use_sr_metadata(self, value):
        self._data_local['sub_use_sr_metadata'] = value

    @property
    def notify_list(self):
        return self._data_local['notify_list']

    @notify_list.setter
    def notify_list(self, value):
        self._data_local['notify_list'] = value

    @property
    def search_delay(self):
        return self._data_local['search_delay']

    @search_delay.setter
    def search_delay(self, value):
        self._data_local['search_delay'] = value

    @property
    def scene_exceptions(self):
        if self._data_local['scene_exceptions']:
            return list(filter(None, self._data_local['scene_exceptions'].split(',')))
        return []

    @scene_exceptions.setter
    def scene_exceptions(self, value):
        self._data_local['scene_exceptions'] = ','.join(value)

    @property
    def last_update(self):
        return self._data_local['last_update']

    @last_update.setter
    def last_update(self, value):
        self._data_local['last_update'] = value

    @property
    def last_refresh(self):
        return self._data_local['last_refresh']

    @last_refresh.setter
    def last_refresh(self, value):
        self._data_local['last_refresh'] = value

    @property
    def last_backlog_search(self):
        return self._data_local['last_backlog_search']

    @last_backlog_search.setter
    def last_backlog_search(self, value):
        self._data_local['last_backlog_search'] = value

    @property
    def last_proper_search(self):
        return self._data_local['last_proper_search']

    @last_proper_search.setter
    def last_proper_search(self, value):
        self._data_local['last_proper_search'] = value

    @property
    def last_scene_exceptions_refresh(self):
        return self._data_local['last_scene_exceptions_refresh']

    @last_scene_exceptions_refresh.setter
    def last_scene_exceptions_refresh(self, value):
        self._data_local['last_scene_exceptions_refresh'] = value

    @property
    def last_xem_refresh(self):
        return self._data_local['last_xem_refresh']

    @last_xem_refresh.setter
    def last_xem_refresh(self, value):
        self._data_local['last_xem_refresh'] = value

    @property
    def series_provider(self):
        return sickrage.app.series_providers[self.series_provider_id]

    @property
    def episodes(self):
        if not self._episodes:
            with sickrage.app.main_db.session() as session:
                for x in session.query(MainDB.TVEpisode).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id):
                    self._episodes[x.episode_id] = TVEpisode(series_id=x.series_id, series_provider_id=x.series_provider_id, season=x.season, episode=x.episode)
        return list(self._episodes.values())

    @property
    def imdb_info(self):
        with sickrage.app.main_db.session() as session:
            return session.query(MainDB.IMDbInfo).filter_by(series_id=self.series_id).one_or_none()

    @property
    def is_anime(self):
        return int(self.anime) > 0

    @property
    def airs_next(self):
        _airs_next = datetime.date.min

        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.season > 0,
                MainDB.TVEpisode.airdate >= datetime.date.today(),
                MainDB.TVEpisode.status.in_([EpisodeStatus.UNAIRED, EpisodeStatus.WANTED])
            ).order_by(
                MainDB.TVEpisode.airdate
            ).first()

            if query:
                _airs_next = query.airdate

        return _airs_next

    @property
    def airs_prev(self):
        _airs_prev = datetime.date.min

        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.season > 0,
                MainDB.TVEpisode.airdate < datetime.date.today(),
                MainDB.TVEpisode.status != EpisodeStatus.UNAIRED
            ).order_by(
                sqlalchemy.desc(MainDB.TVEpisode.airdate)
            ).first()

            if query:
                _airs_prev = query.airdate

        return _airs_prev

    @property
    def episodes_unaired(self):
        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode.season, MainDB.TVEpisode.status
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.status == EpisodeStatus.UNAIRED
            )

            if not sickrage.app.config.gui.display_show_specials:
                query = query.filter(MainDB.TVEpisode.season > 0)

        return query.count()

    @property
    def episodes_snatched(self):
        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode.season, MainDB.TVEpisode.status
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.status.in_(flatten([EpisodeStatus.composites(EpisodeStatus.SNATCHED), EpisodeStatus.composites(EpisodeStatus.SNATCHED_BEST),
                                                     EpisodeStatus.composites(EpisodeStatus.SNATCHED_PROPER)]))
            )

            if not sickrage.app.config.gui.display_show_specials:
                query = query.filter(MainDB.TVEpisode.season > 0)

        return query.count()

    @property
    def episodes_downloaded(self):
        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode.season, MainDB.TVEpisode.status
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.status.in_(flatten([EpisodeStatus.composites(EpisodeStatus.DOWNLOADED), EpisodeStatus.composites(EpisodeStatus.ARCHIVED)]))
            )

            if not sickrage.app.config.gui.display_show_specials:
                query = query.filter(MainDB.TVEpisode.season > 0)

        return query.count()

    @property
    def episodes_special(self):
        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode.season
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.season == 0
            )

        return query.count()

    @property
    def episodes_total(self):
        with sickrage.app.main_db.session() as session:
            query = session.query(
                MainDB.TVEpisode.season, MainDB.TVEpisode.status
            ).filter_by(
                series_id=self.series_id, series_provider_id=self.series_provider_id
            ).filter(
                MainDB.TVEpisode.status != EpisodeStatus.UNAIRED
            )

            if not sickrage.app.config.gui.display_show_specials:
                query = query.filter(MainDB.TVEpisode.season > 0)

        return query.count()

    @property
    def new_episodes(self):
        cur_date = datetime.date.today()
        cur_date += datetime.timedelta(days=1)
        cur_time = datetime.datetime.now(sickrage.app.tz)

        new_episodes = []
        for episode_object in self.episodes:
            if episode_object.status != EpisodeStatus.UNAIRED or episode_object.season == 0 or not episode_object.airdate > datetime.date.min:
                continue

            air_date = episode_object.airdate
            air_date += datetime.timedelta(days=episode_object.show.search_delay)
            if not cur_date >= air_date:
                continue

            if episode_object.show.airs and episode_object.show.network:
                # This is how you assure it is always converted to local time
                air_time = sickrage.app.tz_updater.parse_date_time(episode_object.airdate,
                                                                   episode_object.show.airs,
                                                                   episode_object.show.network).astimezone(sickrage.app.tz)

                # filter out any episodes that haven't started airing yet,
                # but set them to the default status while they are airing
                # so they are snatched faster
                if air_time > cur_time:
                    continue

            new_episodes += [episode_object]

        return new_episodes

    @property
    def total_size(self):
        _total_size = 0
        _related_episodes = set()
        for episode_object in self.episodes:
            [_related_episodes.add(related_episode.episode_id) for related_episode in episode_object.related_episodes]
            if episode_object.episode_id not in _related_episodes:
                _total_size += episode_object.file_size
        return _total_size

    @property
    def network_logo_name(self):
        return unidecode(self.network).lower()

    @property
    def release_groups(self):
        if self.is_anime:
            return BlackAndWhiteList(self.series_id, self.series_provider_id)

    @property
    def poster(self):
        return series_image(self.series_id, self.series_provider_id, SeriesImageType.POSTER).url

    @property
    def banner(self):
        return series_image(self.series_id, self.series_provider_id, SeriesImageType.BANNER).url

    @property
    def allowed_qualities(self):
        allowed_qualities, __ = Quality.split_quality(self.quality)
        return allowed_qualities

    @property
    def preferred_qualities(self):
        __, preferred_qualities = Quality.split_quality(self.quality)
        return preferred_qualities

    @property
    def show_queue_status(self):
        if sickrage.app.show_queue.is_being_added(self.series_id):
            return {
                'action': 'ADD',
                'message': _('This show is in the process of being downloaded - the info below is incomplete.')
            }
        elif sickrage.app.show_queue.is_being_removed(self.series_id):
            return {
                'action': 'REMOVE',
                'message': _('This show is in the process of being removed.')
            }
        elif sickrage.app.show_queue.is_being_updated(self.series_id):
            return {
                'action': 'UPDATE',
                'message': _('The information on this page is in the process of being updated.')
            }
        elif sickrage.app.show_queue.is_being_refreshed(self.series_id):
            return {
                'action': 'REFRESH',
                'message': _('The episodes below are currently being refreshed from disk')
            }
        elif sickrage.app.show_queue.is_being_subtitled(self.series_id):
            return {
                'action': 'SUBTITLE',
                'message': _('Currently downloading subtitles for this show')
            }
        elif sickrage.app.show_queue.is_queued_to_refresh(self.series_id):
            return {
                'action': 'REFRESH_QUEUE',
                'message': _('This show is queued to be refreshed.')
            }
        elif sickrage.app.show_queue.is_queued_to_update(self.series_id):
            return {
                'action': 'UPDATE_QUEUE',
                'message': _('This show is queued and awaiting an update.')
            }
        elif sickrage.app.show_queue.is_queued_to_subtitle(self.series_id):
            return {
                'action': 'SUBTITLE_QUEUE',
                'message': _('This show is queued and awaiting subtitles download.')
            }

        return {}

    @property
    def is_loading(self):
        return self.show_queue_status.get('action') == 'ADD'

    @property
    def is_removing(self):
        return self.show_queue_status.get('action') == 'REMOVE'

    def save(self):
        with self.lock, sickrage.app.main_db.session() as session:
            sickrage.app.log.debug("{0:d}: Saving to database: {1}".format(self.series_id, self.name))

            try:
                query = session.query(MainDB.TVShow).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id).one()
                query.update(**self._data_local)
            except orm.exc.NoResultFound:
                session.add(MainDB.TVShow(**self._data_local))
            finally:
                session.commit()

    def delete(self):
        with self.lock, sickrage.app.main_db.session() as session:
            session.query(MainDB.TVShow).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id).delete()
            session.commit()

    def flush_episodes(self):
        self._episodes.clear()

    def load_from_series_provider(self, cache=True):
        sickrage.app.log.debug(str(self.series_id) + ": Loading show info from " + self.series_provider.name)

        series_provider_language = self.lang or sickrage.app.config.general.series_provider_default_language

        myEp = self.series_provider.search(self.series_id, language=series_provider_language, enable_cache=cache)
        if not myEp:
            raise SeriesProviderException

        try:
            self.name = myEp['seriesname'].strip()
        except AttributeError:
            raise SeriesProviderAttributeNotFound("Found %s, but attribute 'seriesname' was empty." % self.series_id)

        self.overview = safe_getattr(myEp, 'overview', self.overview)
        self.classification = safe_getattr(myEp, 'classification', self.classification)
        self.genre = safe_getattr(myEp, 'genre', self.genre)
        self.network = safe_getattr(myEp, 'network', self.network)
        self.runtime = try_int(safe_getattr(myEp, 'runtime', self.runtime))
        self.imdb_id = safe_getattr(myEp, 'imdbid', self.imdb_id)

        try:
            self.airs = (safe_getattr(myEp, 'airsdayofweek') + " " + safe_getattr(myEp, 'airstime')).strip()
        except:
            self.airs = ''

        try:
            self.startyear = try_int(str(safe_getattr(myEp, 'firstaired') or datetime.date.min).split('-')[0])
        except:
            self.startyear = 0

        self.status = safe_getattr(myEp, 'status', self.status)

        self.save()

    def load_episodes_from_series_provider(self, cache=True):
        scanned_eps = {}

        series_provider_language = self.lang or sickrage.app.config.general.series_provider_default_language

        sickrage.app.log.debug(str(self.series_id) + ": Loading all episodes from " + self.series_provider.name + "..")

        # flush episodes from cache so we can reload from database
        self.flush_episodes()

        series_provider_data = self.series_provider.search(self.series_id, language=series_provider_language, enable_cache=cache)
        if not series_provider_data:
            raise SeriesProviderException

        for season in series_provider_data:
            scanned_eps[season] = {}
            for episode in series_provider_data[season]:
                # need some examples of wtf episode 0 means to decide if we want it or not
                if episode == 0:
                    continue

                try:
                    episode_obj = self.get_episode(season, episode)
                except EpisodeNotFoundException:
                    continue
                else:
                    try:
                        episode_obj.load_from_series_provider(season, episode)
                        episode_obj.save()
                    except EpisodeDeletedException:
                        sickrage.app.log.info("The episode was deleted, skipping the rest of the load")
                        continue

                scanned_eps[season][episode] = True

        # Done updating save last update date
        self.last_update = datetime.datetime.now()

        self.save()

        return scanned_eps

    def get_episode(self, season=None, episode=None, absolute_number=None, location=None, no_create=False):
        try:
            if season is None and episode is None and absolute_number is not None:
                with sickrage.app.main_db.session() as session:
                    query = session.query(MainDB.TVEpisode).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id,
                                                                      absolute_number=absolute_number).one()
                    sickrage.app.log.debug("Found episode by absolute_number %s which is S%02dE%02d" % (absolute_number, query.season, query.episode))
                    season = query.season
                    episode = query.episode

            for tv_episode in self.episodes:
                if tv_episode.season == season and tv_episode.episode == episode:
                    return tv_episode
            else:
                if no_create:
                    return None

                tv_episode = TVEpisode(series_id=self.series_id,
                                       series_provider_id=self.series_provider_id,
                                       season=season,
                                       episode=episode,
                                       location=location or '')

                self._episodes[tv_episode.episode_id] = tv_episode
                return tv_episode
        except orm.exc.MultipleResultsFound:
            if absolute_number is not None:
                sickrage.app.log.debug("Multiple entries for absolute number: " + str(absolute_number) + " in show: " + self.name + " found ")
            raise MultipleEpisodesInDatabaseException
        except orm.exc.NoResultFound:
            if absolute_number is not None:
                sickrage.app.log.debug("No entries for absolute number: " + str(absolute_number) + " in show: " + self.name + " found.")
            raise EpisodeNotFoundException

    def delete_episode(self, season, episode, full=False):
        episode_object = self.get_episode(season, episode, no_create=True)
        if not episode_object:
            return

        data = sickrage.app.notification_providers['trakt'].trakt_episode_data_generate([(episode_object.season, episode_object.episode)])
        if sickrage.app.config.trakt.enable and sickrage.app.config.trakt.sync_watchlist and data:
            sickrage.app.log.debug("Deleting episode from Trakt")
            sickrage.app.notification_providers['trakt'].update_watchlist(self, data_episode=data, update="remove")

        if full and os.path.isfile(episode_object.location):
            sickrage.app.log.info('Attempt to delete episode file %s' % episode_object.location)

            try:
                os.remove(episode_object.location)
            except OSError as e:
                sickrage.app.log.warning('Unable to delete episode file %s: %s / %s' % (episode_object.location, repr(e), str(e)))

        # delete episode from show episode cache
        sickrage.app.log.debug("Deleting %s S%02dE%02d from the shows episode cache" % (self.name, episode_object.season or 0, episode_object.episode or 0))
        try:
            del self._episodes[episode_object.episode_id]
        except KeyError:
            pass

        # delete episode from database
        sickrage.app.log.debug("Deleting %s S%02dE%02d from the DB" % (self.name, episode_object.season or 0, episode_object.episode or 0))
        episode_object.delete()

        raise EpisodeDeletedException()

    def write_show_nfo(self, force=False):
        result = False

        if not os.path.isdir(self.location):
            sickrage.app.log.info(str(self.series_id) + ": Show dir doesn't exist, skipping NFO generation")
            return False

        sickrage.app.log.debug(str(self.series_id) + ": Writing NFOs for show")
        for cur_provider in sickrage.app.metadata_providers.values():
            result = cur_provider.create_show_metadata(self, force) or result

        return result

    def write_metadata(self, show_only=False, force=False):
        if not os.path.isdir(self.location):
            sickrage.app.log.info(str(self.series_id) + ": Show dir doesn't exist, skipping NFO generation")
            return

        self.get_images()

        self.write_show_nfo(force)

        if not show_only:
            self.write_episode_nfos(force)
            self.update_episode_video_metadata()

    def write_episode_nfos(self, force=False):
        if not os.path.isdir(self.location):
            sickrage.app.log.info(str(self.series_id) + ": Show dir doesn't exist, skipping NFO generation")
            return

        sickrage.app.log.debug(str(self.series_id) + ": Writing NFOs for all episodes")

        for episode_obj in self.episodes:
            if episode_obj.location == '':
                continue

            sickrage.app.log.debug(str(self.series_id) + ": Retrieving/creating episode S%02dE%02d" % (episode_obj.season or 0, episode_obj.episode or 0))

            episode_obj.create_meta_files(force)

    def update_episode_video_metadata(self):
        if not os.path.isdir(self.location):
            sickrage.app.log.info(str(self.series_id) + ": Show dir doesn't exist, skipping video metadata updating")
            return

        sickrage.app.log.debug(str(self.series_id) + ": Updating video metadata for all episodes")

        for episode_obj in self.episodes:
            if episode_obj.location == '':
                continue

            sickrage.app.log.debug(str(self.series_id) + ": Updating video metadata for episode S%02dE%02d" % (episode_obj.season, episode_obj.episode))

            episode_obj.update_video_metadata()

    # find all media files in the show folder and create episodes for as many as possible
    def load_episodes_from_dir(self):
        from sickrage.core.nameparser import NameParser, InvalidNameException, InvalidShowException

        if not os.path.isdir(self.location):
            sickrage.app.log.debug(str(self.series_id) + ": Show dir doesn't exist, not loading episodes from disk")
            return

        sickrage.app.log.debug(str(self.series_id) + ": Loading all episodes from the show directory " + self.location)

        # get file list
        media_files = list_media_files(self.location)

        # create TVEpisodes from each media file (if possible)
        for mediaFile in media_files:
            curEpisode = None

            sickrage.app.log.debug(str(self.series_id) + ": Creating episode from " + mediaFile)
            try:
                curEpisode = self.make_ep_from_file(os.path.join(self.location, mediaFile))
            except (ShowNotFoundException, EpisodeNotFoundException) as e:
                sickrage.app.log.warning("Episode " + mediaFile + " returned an exception: {}".format(e))
            except EpisodeDeletedException:
                sickrage.app.log.debug("The episode deleted itself when I tried making an object for it")

            # skip to next episode?
            if not curEpisode:
                continue

            # see if we should save the release name in the db
            ep_file_name = os.path.basename(curEpisode.location)
            ep_file_name = os.path.splitext(ep_file_name)[0]

            try:
                parse_result = NameParser(False, series_id=self.series_id, series_provider_id=self.series_provider_id).parse(ep_file_name)
            except (InvalidNameException, InvalidShowException):
                parse_result = None

            if ' ' not in ep_file_name and parse_result and parse_result.release_group:
                sickrage.app.log.debug("Name " + ep_file_name + " gave release group of " + parse_result.release_group + ", seems valid")
                curEpisode.release_name = ep_file_name
                self.save()

            # store the reference in the show
            if self.subtitles and sickrage.app.config.subtitles.enable:
                try:
                    curEpisode.refresh_subtitles()
                except Exception:
                    sickrage.app.log.error("%s: Could not refresh subtitles" % self.series_id)
                    sickrage.app.log.debug(traceback.format_exc())

    def load_imdb_info(self):
        imdb_info_mapper = {
            'imdbvotes': 'votes',
            'imdbrating': 'rating',
            'totalseasons': 'seasons',
            'imdbid': 'imdb_id'
        }

        if not re.search(r'^tt\d+$', self.imdb_id) and self.name:
            resp = sickrage.app.api.imdb.search_by_imdb_title(self.name)
            if not resp:
                resp = {}

            for x in resp.get('Search', []):
                try:
                    if int(x.get('Year'), 0) == self.startyear and x.get('Title') in self.name:
                        if re.search(r'^tt\d+$', x.get('imdbID', '')):
                            self.imdb_id = x.get('imdbID')
                            self.save()
                            break
                except Exception:
                    continue

        if re.search(r'^tt\d+$', self.imdb_id):
            sickrage.app.log.debug(str(self.series_id) + ": Obtaining IMDb info")

            imdb_info = sickrage.app.api.imdb.search_by_imdb_id(self.imdb_id)
            if not imdb_info:
                sickrage.app.log.debug(str(self.series_id) + ': Unable to obtain IMDb info')
                return

            imdb_info = dict((k.lower(), v) for k, v in imdb_info.items())
            for column in imdb_info.copy():
                if column in imdb_info_mapper:
                    imdb_info[imdb_info_mapper[column]] = imdb_info[column]

                if column not in MainDB.IMDbInfo.__table__.columns.keys():
                    del imdb_info[column]

            if not all([imdb_info.get('imdb_id'), imdb_info.get('votes'), imdb_info.get('rating'), imdb_info.get('genre')]):
                sickrage.app.log.debug(str(self.series_id) + ': IMDb info obtained does not meet our requirements')
                return

            sickrage.app.log.debug(str(self.series_id) + ": Obtained IMDb info ->" + str(imdb_info))

            # save imdb info to database
            imdb_info.update({
                'series_id': self.series_id,
                'last_update': datetime.datetime.now()
            })

            with sickrage.app.main_db.session() as session:
                try:
                    dbData = session.query(MainDB.IMDbInfo).filter_by(series_id=self.series_id).one()
                    dbData.update(**imdb_info)
                except orm.exc.NoResultFound:
                    session.add(MainDB.IMDbInfo(**imdb_info))
                finally:
                    self.save()

    def get_images(self, fanart=None, poster=None):
        fanart_result = poster_result = banner_result = False
        season_posters_result = season_banners_result = season_all_poster_result = season_all_banner_result = False

        for cur_provider in sickrage.app.metadata_providers.values():
            fanart_result = cur_provider.create_fanart(self) or fanart_result
            poster_result = cur_provider.create_poster(self) or poster_result
            banner_result = cur_provider.create_banner(self) or banner_result

            season_posters_result = cur_provider.create_season_posters(self) or season_posters_result
            season_banners_result = cur_provider.create_season_banners(self) or season_banners_result
            season_all_poster_result = cur_provider.create_season_all_poster(self) or season_all_poster_result
            season_all_banner_result = cur_provider.create_season_all_banner(self) or season_all_banner_result

        return fanart_result or poster_result or banner_result or season_posters_result or season_banners_result or season_all_poster_result or season_all_banner_result

    def make_ep_from_file(self, filename):
        from sickrage.core.nameparser import NameParser, InvalidNameException, InvalidShowException

        if not os.path.isfile(filename):
            sickrage.app.log.info(str(self.series_id) + ": That isn't even a real file dude... " + filename)
            return None

        sickrage.app.log.debug(str(self.series_id) + ": Creating episode object from " + filename)

        try:
            parse_result = NameParser(validate_show=False).parse(filename, skip_scene_detection=True)
        except InvalidNameException:
            sickrage.app.log.debug("Unable to parse the filename " + filename + " into a valid episode")
            return None
        except InvalidShowException:
            sickrage.app.log.debug("Unable to parse the filename " + filename + " into a valid show")
            return None

        if not len(parse_result.episode_numbers):
            sickrage.app.log.info("parse_result: " + str(parse_result))
            sickrage.app.log.warning("No episode number found in " + filename + ", ignoring it")
            return None

        # for now lets assume that any episode in the show dir belongs to that show
        season = parse_result.season_number if parse_result.season_number is not None else 1
        root_ep = None

        for curEpNum in parse_result.episode_numbers:
            episode = int(curEpNum)

            sickrage.app.log.debug("%s: %s parsed to %s S%02dE%02d" % (self.series_id, filename, self.name, int(season or 0), int(episode or 0)))

            check_quality_again = False

            try:
                episode_obj = self.get_episode(season, episode, location=filename)
            except EpisodeNotFoundException:
                sickrage.app.log.warning("{}: Unable to figure out what this file is, skipping {}".format(self.series_id, filename))
                continue

            # if there is a new file associated with this ep then re-check the quality
            if episode_obj.location and os.path.normpath(episode_obj.location) != os.path.normpath(filename):
                sickrage.app.log.debug("The old episode had a different file associated with it, I will re-check "
                                       "the quality based on the new filename " + filename)
                check_quality_again = True

            # if the sizes are the same then it's probably the same file
            old_size = episode_obj.file_size
            episode_obj.location = filename
            same_file = old_size and episode_obj.file_size == old_size
            episode_obj.checkForMetaFiles()

            if root_ep is None:
                root_ep = episode_obj
            else:
                if episode_obj not in root_ep.related_episodes:
                    root_ep.related_episodes.append(episode_obj)

            # if it's a new file then
            if not same_file:
                episode_obj.release_name = ''

            # if they replace a file on me I'll make some attempt at re-checking the quality unless I know it's the
            # same file
            if check_quality_again and not same_file:
                new_quality = Quality.name_quality(filename, self.is_anime)
                sickrage.app.log.debug("Since this file has been renamed")

                episode_obj.status = Quality.composite_status(EpisodeStatus.DOWNLOADED, new_quality)

            # check for status/quality changes as long as it's a new file
            elif not same_file and is_media_file(
                    filename) and episode_obj.status not in flatten([EpisodeStatus.composites(EpisodeStatus.DOWNLOADED),
                                                                     EpisodeStatus.composites(EpisodeStatus.ARCHIVED), EpisodeStatus.IGNORED]):
                old_status, old_quality = Quality.split_composite_status(episode_obj.status)
                new_quality = Quality.name_quality(filename, self.is_anime)

                new_status = None

                # if it was snatched and now exists then set the status correctly
                if old_status == EpisodeStatus.SNATCHED and old_quality <= new_quality:
                    sickrage.app.log.debug(
                        "STATUS: this ep used to be snatched with quality " + old_quality.display_name +
                        " but a file exists with quality " + new_quality.display_name +
                        " so I'm setting the status to DOWNLOADED")
                    new_status = EpisodeStatus.DOWNLOADED

                # if it was snatched proper and we found a higher quality one then allow the status change
                elif old_status == EpisodeStatus.SNATCHED_PROPER and old_quality < new_quality:
                    sickrage.app.log.debug(
                        "STATUS: this ep used to be snatched proper with quality " + old_quality.display_name +
                        " but a file exists with quality " + new_quality.display_name +
                        " so I'm setting the status to DOWNLOADED")
                    new_status = EpisodeStatus.DOWNLOADED

                elif old_status not in (EpisodeStatus.SNATCHED, EpisodeStatus.SNATCHED_PROPER):
                    new_status = EpisodeStatus.DOWNLOADED

                if new_status is not None:
                    sickrage.app.log.debug(
                        "STATUS: we have an associated file, so setting the status from " + str(
                            episode_obj.status) + " to DOWNLOADED/" + str(Quality.status_from_name(filename, anime=self.is_anime)))
                    episode_obj.status = Quality.composite_status(new_status, new_quality)

            # save episode data to database
            episode_obj.save()

        # creating metafiles on the root should be good enough
        if root_ep:
            root_ep.create_meta_files()

        # save show data to database
        self.save()

        return root_ep

    def delete_show(self, full=False):
        # choose delete or trash action
        action = ('delete', 'trash')[sickrage.app.config.general.trash_remove_show]

        # remove from database
        with sickrage.app.main_db.session() as session:
            series = session.query(MainDB.TVShow).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id).one()
            session.delete(series)
            session.commit()

        # remove episodes from show episode cache
        self.flush_episodes()

        # remove from show cache if found
        try:
            del sickrage.app.shows[(self.series_id, self.series_provider_id)]
        except KeyError:
            pass

        # clear the cache
        image_cache_dir = os.path.join(sickrage.app.cache_dir, 'images')
        for cache_file in glob.glob(os.path.join(image_cache_dir, str(self.series_id) + '.*')):
            sickrage.app.log.info('Attempt to %s cache file %s' % (action, cache_file))
            try:
                if sickrage.app.config.general.trash_remove_show:
                    send2trash.send2trash(cache_file)
                else:
                    os.remove(cache_file)
            except OSError as e:
                sickrage.app.log.warning('Unable to %s %s: %s / %s' % (action, cache_file, repr(e), str(e)))

        # remove entire show folder
        if full:
            try:
                if os.path.isdir(self.location):
                    sickrage.app.log.info('Attempt to %s show folder %s' % (action, self.location))

                    # check first the read-only attribute
                    file_attribute = os.stat(self.location)[0]
                    if not file_attribute & stat.S_IWRITE:
                        # File is read-only, so make it writeable
                        sickrage.app.log.debug(
                            'Attempting to make writeable the read only folder %s' % self.location)
                        try:
                            os.chmod(self.location, stat.S_IWRITE)
                        except Exception:
                            sickrage.app.log.warning('Unable to change permissions of %s' % self.location)

                    if sickrage.app.config.general.trash_remove_show:
                        send2trash.send2trash(self.location)
                    else:
                        shutil.rmtree(self.location)

                    sickrage.app.log.info('%s show folder %s' %
                                          (('Deleted', 'Trashed')[sickrage.app.config.general.trash_remove_show],
                                           self.location))
            except OSError as e:
                sickrage.app.log.warning('Unable to %s %s: %s / %s' % (action, self.location, repr(e), str(e)))

        if sickrage.app.config.trakt.enable and sickrage.app.config.trakt.sync_watchlist:
            sickrage.app.log.debug("Removing show: {}, {} from watchlist".format(self.series_id, self.name))
            sickrage.app.notification_providers['trakt'].update_watchlist(self, update="remove")

    def populate_cache(self, force=False):
        sickrage.app.log.debug("Checking & filling cache for show " + self.name)
        ImageCache().fill_cache(self, force)

    def refresh_dir(self):
        # make sure the show dir is where we think it is unless dirs are created on the fly
        if not os.path.isdir(self.location) and not sickrage.app.config.general.create_missing_show_dirs:
            return False

        # load from dir
        try:
            self.load_episodes_from_dir()
        except Exception as e:
            sickrage.app.log.debug("Error searching dir for episodes: {}".format(e))
            sickrage.app.log.debug(traceback.format_exc())

        # run through all locations from DB, check that they exist
        sickrage.app.log.debug(str(self.series_id) + ": Loading all episodes with a location from the database")

        for curEp in self.episodes:
            if curEp.location == '':
                continue

            curLoc = os.path.normpath(curEp.location)
            season = int(curEp.season)
            episode = int(curEp.episode)

            # if the path doesn't exist or if it's not in our show dir
            if not os.path.isfile(curLoc) or not os.path.normpath(curLoc).startswith(os.path.normpath(self.location)):
                # check if downloaded files still exist, update our data if this has changed
                if not sickrage.app.config.general.skip_removed_files:
                    # if it used to have a file associated with it and it doesn't anymore then set it to
                    # EP_DEFAULT_DELETED_STATUS
                    if curEp.location and curEp.status in EpisodeStatus.composites(EpisodeStatus.DOWNLOADED):
                        if sickrage.app.config.general.ep_default_deleted_status == EpisodeStatus.ARCHIVED:
                            __, oldQuality = Quality.split_composite_status(curEp.status)
                            new_status = Quality.composite_status(EpisodeStatus.ARCHIVED, oldQuality)
                        else:
                            new_status = sickrage.app.config.general.ep_default_deleted_status

                        sickrage.app.log.debug("%s: Location for S%02dE%02d doesn't exist, "
                                               "removing it and changing our status to %s" % (self.series_id,
                                                                                              season or 0, episode or 0,
                                                                                              new_status.display_name))

                        curEp.status = new_status
                        curEp.subtitles = ''
                        curEp.subtitles_searchcount = 0
                        curEp.subtitles_lastsearch = datetime.datetime.min

                    curEp.location = ''
                    curEp.hasnfo = False
                    curEp.hastbn = False
                    curEp.release_name = ''
            else:
                if curEp.status in EpisodeStatus.composites(EpisodeStatus.ARCHIVED):
                    __, oldQuality = Quality.split_composite_status(curEp.status)
                    curEp.status = Quality.composite_status(EpisodeStatus.DOWNLOADED, oldQuality)

                # the file exists, set its modify file stamp
                if sickrage.app.config.general.airdate_episodes:
                    curEp.airdate_modify_stamp()

            # save episode to database
            curEp.save()

    def download_subtitles(self):
        if not os.path.isdir(self.location):
            sickrage.app.log.debug(str(self.series_id) + ": Show dir doesn't exist, can't download subtitles")
            return

        sickrage.app.log.debug("%s: Downloading subtitles" % self.series_id)

        try:
            for episode in self.episodes:
                episode.download_subtitles()
        except Exception:
            sickrage.app.log.error("%s: Error occurred when downloading subtitles for %s" % (self.series_id, self.name))

    def qualitiesToString(self, qualities=None):
        if qualities is None:
            qualities = []

        result = ''
        for quality in qualities:
            if quality in Qualities:
                result += quality.display_name + ', '

        result = re.sub(', $', '', result)

        if not len(result):
            result = 'None'

        return result

    def want_episode(self, season, episode, quality, manualSearch=False, downCurQuality=False):
        try:
            episode_object = self.get_episode(season, episode)
        except EpisodeNotFoundException:
            sickrage.app.log.debug("Unable to find a matching episode in database, ignoring found episode")
            return False

        sickrage.app.log.debug("Checking if found episode %s S%02dE%02d is wanted at quality %s" % (
            self.name, episode_object.season or 0, episode_object.episode or 0, quality.display_name))

        # if the quality isn't one we want under any circumstances then just say no
        any_qualities, best_qualities = Quality.split_quality(self.quality)
        sickrage.app.log.debug("Any, Best = [{}] [{}] Found = [{}]".format(
            self.qualitiesToString(any_qualities),
            self.qualitiesToString(best_qualities),
            self.qualitiesToString([quality]))
        )

        if quality not in any_qualities + best_qualities or quality is EpisodeStatus.UNKNOWN:
            sickrage.app.log.debug("Don't want this quality, ignoring found episode")
            return False

        ep_status, ep_quality = Quality.split_composite_status(episode_object.status)

        sickrage.app.log.debug(f"Existing episode status: {ep_status.display_name}")

        # if we know we don't want it then just say no
        if ep_status in flatten(
                [EpisodeStatus.composites(EpisodeStatus.ARCHIVED), EpisodeStatus.UNAIRED, EpisodeStatus.SKIPPED, EpisodeStatus.IGNORED]) and not manualSearch:
            sickrage.app.log.debug("Existing episode status is unaired/skipped/ignored/archived, ignoring found episode")
            return False

        # if it's one of these then we want it as long as it's in our allowed initial qualities
        if ep_status == EpisodeStatus.WANTED:
            sickrage.app.log.debug("Existing episode status is WANTED, getting found episode")
            return True
        elif manualSearch:
            if (downCurQuality and quality >= ep_quality) or (not downCurQuality and quality > ep_quality):
                sickrage.app.log.debug("Usually ignoring found episode, but forced search allows the quality, getting found episode")
                return True

        # if we are re-downloading then we only want it if it's in our bestQualities list and better than what we
        # have, or we only have one bestQuality and we do not have that quality yet
        if ep_status in flatten([EpisodeStatus.composites(EpisodeStatus.DOWNLOADED), EpisodeStatus.composites(EpisodeStatus.SNATCHED),
                                 EpisodeStatus.composites(EpisodeStatus.SNATCHED_PROPER)]) and quality in best_qualities and (
                quality > ep_quality or ep_quality not in best_qualities):
            sickrage.app.log.debug("Episode already exists but the found episode quality is wanted more, getting found episode")
            return True
        elif ep_quality == EpisodeStatus.UNKNOWN and manualSearch:
            sickrage.app.log.debug("Episode already exists but quality is Unknown, getting found episode")
            return True
        else:
            sickrage.app.log.debug("Episode already exists and the found episode has same/lower quality, ignoring found episode")

        sickrage.app.log.debug("None of the conditions were met, ignoring found episode")
        return False

    def get_all_episodes_from_absolute_number(self, absolute_numbers):
        episodes = []
        season = None

        if len(absolute_numbers):
            for absolute_number in absolute_numbers:
                try:
                    ep = self.get_episode(absolute_number=absolute_number)
                    episodes.append(ep.episode)
                    season = ep.season
                except (EpisodeNotFoundException, MultipleEpisodesInDatabaseException):
                    continue

        return season, episodes

    def retrieve_scene_exceptions(self, get_anidb=True, force=False):
        """
        Looks up the exceptions on SR API.
        """

        max_refresh_age_secs = 86400  # 1 day

        if not datetime.datetime.now() > (self.last_scene_exceptions_refresh + datetime.timedelta(seconds=max_refresh_age_secs)) and not force:
            return

        try:
            sickrage.app.log.debug("Retrieving scene exceptions from SiCKRAGE API for show: {}".format(self.name))

            scene_exceptions = sickrage.app.api.scene_exceptions.search_by_id(self.series_id)
            if not scene_exceptions or 'data' not in scene_exceptions:
                sickrage.app.log.debug("No scene exceptions found from SiCKRAGE API for show: {}".format(self.name))
            else:
                self.scene_exceptions = set(self.scene_exceptions + scene_exceptions['data']['exceptions'].split(','))

            if get_anidb and self.is_anime and self.series_provider_id == SeriesProviderID.THETVDB:
                try:
                    sickrage.app.log.info("Retrieving AniDB scene exceptions for show: {}".format(self.name))
                    anime = Anime(None, name=self.name, tvdbid=self.series_id, autoCorrectName=True)
                    if anime and anime.name != self.name:
                        anidb_scene_exceptions = ['{}|-1'.format(anime.name)]
                        self.scene_exceptions = set(self.scene_exceptions + anidb_scene_exceptions)
                except Exception:
                    pass

            self.last_scene_exceptions_refresh = datetime.datetime.now()
            self.save()
        except Exception as e:
            sickrage.app.log.debug("Check scene exceptions update failed from SiCKRAGE API for show: {}".format(self.name))

    def get_scene_exception_by_name(self, exception_name):
        for x in self.scene_exceptions:
            if exception_name in x:
                scene_name, scene_season = x.split('|')
                return scene_name, int(scene_season)

    def get_scene_exceptions_by_season(self, season=-1):
        scene_exceptions = []

        for scene_exception in self.scene_exceptions:
            try:
                scene_name, scene_season = scene_exception.split('|')
                if season == int(scene_season):
                    scene_exceptions.append(scene_name)
            except ValueError:
                pass

        return scene_exceptions

    def update_scene_exceptions(self, scene_exceptions, season=-1):
        self.scene_exceptions = set([x + '|{}'.format(season) for x in scene_exceptions])
        self.save()

    def __unicode__(self):
        to_return = ""
        to_return += "series_id: {}\n".format(self.series_id)
        to_return += "series_provider_id: {}\n".format(self.series_provider_id.display_name)
        to_return += "name: {}\n".format(self.name)
        to_return += "location: {}\n".format(self.location)
        if self.network:
            to_return += "network: {}\n".format(self.network)
        if self.airs:
            to_return += "airs: {}\n".format(self.airs)
        to_return += "status: {}\n".format(self.status)
        to_return += "startyear: {}\n".format(self.startyear)
        if self.genre:
            to_return += "genre: {}\n".format(self.genre)
        to_return += "overview: {}\n".format(self.overview)
        to_return += "classification: {}\n".format(self.classification)
        to_return += "runtime: {}\n".format(self.runtime)
        to_return += "quality: {}\n".format(self.quality)
        to_return += "search format: {}\n".format(self.search_format.display_name)
        to_return += "anime: {}\n".format(self.is_anime)
        return to_return

    def to_json(self, episodes=False, progress=False, details=False):
        with sickrage.app.main_db.session() as session:
            series = session.query(MainDB.TVShow).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id).one_or_none()
            json_data = TVShowSchema().dump(series)

            json_data['seriesSlug'] = self.slug

            json_data['isLoading'] = self.is_loading
            json_data['isRemoving'] = self.is_removing

            # images section
            json_data['images'] = {
                'poster': self.poster,
                'banner': self.banner
            }

            # qualities section
            json_data['qualities'] = {
                'allowedQualities': [x.name for x in self.allowed_qualities],
                'preferredQualities': [x.name for x in self.preferred_qualities]
            }

            # show queue status
            json_data['showQueueStatus'] = self.show_queue_status

            if details:
                # imdb info section
                imdb_info = session.query(MainDB.IMDbInfo).filter_by(series_id=self.series_id, imdb_id=self.imdb_id).one_or_none()
                json_data['imdbInfo'] = IMDbInfoSchema().dump(imdb_info)

                # anime section
                blacklist = session.query(MainDB.Blacklist).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id).one_or_none()
                whitelist = session.query(MainDB.Whitelist).filter_by(series_id=self.series_id, series_provider_id=self.series_provider_id).one_or_none()
                json_data['blacklist'] = BlacklistSchema().dump(blacklist)
                json_data['whitelist'] = WhitelistSchema().dump(whitelist)

            # progress section
            if progress:
                episodes_snatched = self.episodes_snatched
                episodes_downloaded = self.episodes_downloaded
                episodes_total = self.episodes_total
                progressbar_percent = int(episodes_downloaded * 100 / episodes_total if episodes_total > 0 else 1)

                progress_text = '?'
                progress_tip = _("no data")
                if episodes_total != 0:
                    progress_text = str(episodes_downloaded)
                    progress_tip = _("Downloaded: ") + str(episodes_downloaded)
                    if episodes_snatched > 0:
                        progress_text = progress_text + "+" + str(episodes_snatched)
                        progress_tip = progress_tip + "&#013;" + _("Snatched: ") + str(episodes_snatched)

                    progress_text = progress_text + " / " + str(episodes_total)
                    progress_tip = progress_tip + "&#013;" + _("Total: ") + str(episodes_total)

                json_data['progress'] = {
                    'text': progress_text,
                    'tip': progress_tip,
                    'percent': progressbar_percent
                }

            # episodes section
            if episodes:
                json_data['episodes'] = {ep.episode_id: ep.to_json() for ep in self.episodes}

            return json_data



"""Initial migration

Revision ID: 10
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '10'
down_revision = '9'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    history = sa.Table('history', meta, autoload=True)

    if not hasattr(history.c, 'release_group'):
        op.add_column('history', sa.Column('release_group', sa.Text, default=''))


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import os
import platform
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
from distutils.version import LooseVersion
from time import sleep

import dirsync as dirsync

import sickrage
from sickrage.core.helpers import backup_app_data
from sickrage.core.websession import WebSession
from sickrage.core.websocket import WebSocketMessage
from sickrage.notification_providers import NotificationProvider


class VersionUpdater(object):
    def __init__(self):
        self.name = "VERSIONUPDATER"
        self.running = False

    @property
    def updater(self):
        # default to source install type
        install_type = SourceUpdateManager()

        if sickrage.install_type() == 'git':
            # GIT install type
            install_type = GitUpdateManager()
        elif sickrage.install_type() == 'windows':
            # PIP install type
            install_type = WindowsUpdateManager()
        elif sickrage.install_type() == 'synology':
            # PIP install type
            install_type = SynologyUpdateManager()
        elif sickrage.install_type() == 'docker':
            # PIP install type
            install_type = DockerUpdateManager()
        elif sickrage.install_type() == 'qnap':
            # PIP install type
            install_type = QnapUpdateManager()
        elif sickrage.install_type() == 'readynas':
            # PIP install type
            install_type = ReadynasUpdateManager()
        elif sickrage.install_type() == 'pip':
            # PIP install type
            install_type = PipUpdateManager()

        return install_type

    @property
    def version(self):
        return self.updater.version

    @property
    def branch(self):
        return self.updater.current_branch

    def task(self, force=False):
        if self.running:
            return

        try:
            self.running = True

            if not self.check_for_update():
                return

            if not sickrage.app.config.general.auto_update and not force:
                return

            if self.updater.manual_update:
                sickrage.app.log.debug("We can't proceed with auto-updating, install type only allows manual updating")
                return

            if sickrage.app.show_updater.running:
                sickrage.app.log.debug("We can't proceed with auto-updating, shows are being updated")
                return

            sickrage.app.log.info("New update found for SiCKRAGE, starting auto-updater ...")
            sickrage.app.alerts.message(_('Updater'), _('New update found for SiCKRAGE, starting auto-updater'))

            if self.update():
                sickrage.app.log.info("Update was successful!")
                sickrage.app.alerts.message(_('Updater'), _('Update was successful'))
                sickrage.app.shutdown(restart=True)
            else:
                sickrage.app.log.info("Update failed!")
                sickrage.app.alerts.error(_('Updater'), _('Update failed!'))
        finally:
            self.running = False

    def backup(self):
        # Do a system backup before update
        sickrage.app.log.info("Config backup in progress...")
        sickrage.app.alerts.message(_('Updater'), _('Config backup in progress...'))

        try:
            backupDir = os.path.join(sickrage.app.data_dir, 'backup')
            if not os.path.isdir(backupDir):
                os.mkdir(backupDir)

            if backup_app_data(backupDir, keep_latest=True):
                sickrage.app.log.info("Config backup successful, updating...")
                sickrage.app.alerts.message(_('Updater'), _('Config backup successful, updating...'))
                return True
            else:
                sickrage.app.log.warning("Config backup failed, aborting update")
                sickrage.app.alerts.error(_('Updater'), _('Config backup failed, aborting update'))
                return False
        except Exception as e:
            sickrage.app.log.warning(f'Update: Config backup failed. Error: {e!r}')
            sickrage.app.alerts.error(_('Updater'), _('Config backup failed, aborting update'))
            return False

    def safe_to_update(self):
        sickrage.app.postprocessor_queue.shutdown()
        sickrage.app.log.debug("Waiting for jobs in post-processor queue to finish before updating")
        sickrage.app.alerts.message(_('Updater'), _("Waiting for jobs in post-processor queue to finish before updating"))

        while sickrage.app.postprocessor_queue.is_busy:
            sleep(1)

        sickrage.app.show_queue.shutdown()
        sickrage.app.log.debug("Waiting for jobs in show queue to finish before updating")
        sickrage.app.alerts.message(_('Updater'), _("Waiting for jobs in show queue to finish before updating"))

        while sickrage.app.show_queue.is_busy:
            sleep(1)

        return True

    def check_for_update(self):
        """
        Checks the internet for a newer version.

        returns: bool, True for new version or False for no new version.
        :param force: forces return value of True
        """

        if sickrage.app.disable_updates:
            return False

        if not self.updater.need_update():
            return False

        self.updater.set_latest_version()

        return True

    def update(self, webui=False):
        # check if updater only allows manual updates
        if self.updater.manual_update:
            return False

        # check for updates
        if not self.updater.need_update():
            return False

        # check if its safe to update
        if not self.safe_to_update():
            return False

        # backup
        if sickrage.app.config.general.backup_on_update and not self.backup():
            return False

        # attempt update
        if self.updater.update():
            # Clean up after update
            to_clean = os.path.join(sickrage.app.cache_dir, 'mako')

            for root, dirs, files in os.walk(to_clean, topdown=False):
                [os.remove(os.path.join(root, name)) for name in files]
                [shutil.rmtree(os.path.join(root, name)) for name in dirs]

            sickrage.app.config.general.view_changelog = True

            if webui:
                WebSocketMessage('redirect', {'url': f'{sickrage.app.config.general.web_root}/home/restart/?pid={sickrage.app.pid}'}).push()

            return True

        if webui:
            sickrage.app.alerts.error(_("Updater"), _("Update wasn't successful, not restarting. Check your log for more information."))


class UpdateManager(object):
    def __init__(self):
        self.manual_update = False

    @property
    def version(self):
        return sickrage.version()

    @property
    def latest_version(self):
        releases = []
        latest_version = None

        try:
            version_url = "https://git.sickrage.ca/SiCKRAGE/sickrage/-/releases.json"
            resp = WebSession().get(version_url).json()

            if self.current_branch == 'develop':
                releases = [x['tag'] for x in resp if 'dev' in x['tag']]
            elif self.current_branch == 'master':
                releases = [x['tag'] for x in resp if 'dev' not in x['tag']]

            if releases:
                latest_version = sorted(releases, key=LooseVersion, reverse=True)[0]
        finally:
            return latest_version or self.version

    @property
    def current_branch(self):
        return ("master", "develop")["dev" in sickrage.version()]

    def need_update(self):
        try:
            latest_version = self.latest_version
            if LooseVersion(self.version) < LooseVersion(latest_version):
                sickrage.app.log.debug(f"SiCKRAGE version upgrade: {self.version} -> {latest_version}")
                return True
        except Exception as e:
            sickrage.app.log.warning(f"Unable to check for updates: {e!r}")
            return False

    def update(self):
        pass

    def set_latest_version(self):
        latest_version = self.latest_version

        if not self.manual_update:
            update_url = f"{sickrage.app.config.general.web_root}/home/update/?pid={sickrage.app.pid}"
            message = _(f'New SiCKRAGE {self.current_branch} {sickrage.install_type()} update available, version {latest_version} &mdash; <a href=\"{update_url}\">Update Now</a>')
        else:
            message = _(f"New SiCKRAGE {self.current_branch} {sickrage.install_type()} update available, version {latest_version}, please manually update!")

        sickrage.app.latest_version_string = message

    @staticmethod
    def _pip_cmd(args, silent=False):
        output = err = None

        cmd = [sys.executable, "-m", "pip"] + args.split()

        try:
            if not silent:
                sickrage.app.log.debug("Executing " + ' '.join(cmd) + " with your shell in " + sickrage.MAIN_DIR)

            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=(sys.platform == 'win32'),
                                 cwd=sickrage.MAIN_DIR)

            output, err = p.communicate()
            exit_status = p.returncode
        except (RuntimeError, OSError):
            sickrage.app.log.info(f"Command {' '.join(cmd)} didn't work")
            exit_status = 1

        if exit_status == 0:
            exit_status = 0
        else:
            exit_status = 1

        if output:
            output = output.decode("utf-8", "ignore").strip() if isinstance(output, bytes) else output.strip()

        return output, err, exit_status

    def upgrade_pip(self):
        output, __, exit_status = self._pip_cmd('install --no-cache-dir -U pip')

        if exit_status != 0:
            __, __, exit_status = self._pip_cmd('install --no-cache-dir --user -U PIP')

        if exit_status == 0:
            return True

        sickrage.app.alerts.error(_('Updater'), _('Failed to update PIP'))

        sickrage.app.log.warning('Unable to update PIP')

        if output:
            output = output.decode("utf-8", "ignore").strip() if isinstance(output, bytes) else output.strip()
            sickrage.app.log.debug(f"PIP CMD OUTPUT: {output}")

    def install_requirements(self, branch):
        requirements_url = f"https://git.sickrage.ca/SiCKRAGE/sickrage/raw/{branch}/requirements.txt"
        requirements_file = tempfile.NamedTemporaryFile(delete=False)

        try:
            requirements_file.write(WebSession().get(requirements_url).content)
            requirements_file.close()
        except Exception:
            requirements_file.close()
            os.unlink(requirements_file.name)
            return False

        output, __, exit_status = self._pip_cmd(f'install --no-deps --no-cache-dir -r {requirements_file.name}')
        if exit_status != 0:
            __, __, exit_status = self._pip_cmd(f'install --no-deps --no-cache-dir --user -r {requirements_file.name}')

        if exit_status == 0:
            requirements_file.close()
            os.unlink(requirements_file.name)
            return True

        sickrage.app.alerts.error(_('Updater'), _('Failed to update requirements'))

        sickrage.app.log.warning('Unable to update requirements')

        if output:
            output = output.decode("utf-8", "ignore").strip() if isinstance(output, bytes) else output.strip()
            sickrage.app.log.debug("PIP CMD OUTPUT: {}".format(output))

        requirements_file.close()
        os.unlink(requirements_file.name)

        return False


class GitUpdateManager(UpdateManager):
    def __init__(self):
        super(GitUpdateManager, self).__init__()
        self.type = "git"
        self._num_commits_behind = 0
        self._num_commits_ahead = 0

    @property
    def version(self):
        """
        Attempts to find the currently installed version of SiCKRAGE.

        Uses git show to get commit version.

        Returns: True for success or False for failure
        """

        output, __, exit_status = self._git_cmd(self._git_path, 'rev-parse HEAD')
        if exit_status == 0 and output:
            cur_commit_hash = output.strip()
            if not re.match('^[a-z0-9]+$', cur_commit_hash):
                sickrage.app.log.error("Output doesn't look like a hash, not using it")
                return False
            return cur_commit_hash

    @property
    def latest_version(self):
        """
        Uses git commands to check if there is a newer version that the provided
        commit hash. If there is a newer version it sets _num_commits_behind.
        """

        # check if branch exists on remote
        if self.current_branch not in self.remote_branches:
            return self.version

        # get all new info from server
        output, __, exit_status = self._git_cmd(self._git_path, 'remote update')
        if not exit_status == 0:
            sickrage.app.log.warning("Unable to contact server, can't check for update")

            if output:
                sickrage.app.log.debug(f'GIT CMD OUTPUT: {output.strip()}')

            return self.version

        # get number of commits behind and ahead (option --count not supported git < 1.7.2)
        output, __, exit_status = self._git_cmd(self._git_path, f'rev-list --left-right origin/{self.current_branch}...HEAD')
        if exit_status == 0 and output:
            try:
                self._num_commits_behind = int(output.count("<"))
                self._num_commits_ahead = int(output.count(">"))
            except Exception:
                sickrage.app.log.debug("Unable to determine number of commits ahead or behind for git install, failed new version check.")
                return self.version

        # get latest commit_hash from remote
        output, __, exit_status = self._git_cmd(self._git_path, f'rev-parse --verify --quiet origin/{self.current_branch}')
        if exit_status == 0 and output:
            return output.strip() or self.version

    @property
    def current_branch(self):
        branch_ref, __, exit_status = self._git_cmd(self._git_path, 'symbolic-ref -q HEAD')
        if exit_status == 0 and branch_ref is not None:
            return branch_ref.strip().replace('refs/heads/', '', 1)
        return ""

    @property
    def remote_branches(self):
        branches, __, exit_status = self._git_cmd(self._git_path, f'ls-remote --heads {sickrage.app.git_remote_url}')
        if exit_status == 0 and branches:
            return re.findall(r'refs/heads/(.*)', branches)

        return []

    @property
    def _git_path(self):
        test_cmd = '--version'

        alternative_git = {
            'windows': 'git',
            'darwin': '/usr/local/git/bin/git'
        }

        main_git = sickrage.app.config.general.git_path or 'git'

        # sickrage.app.log.debug("Checking if we can use git commands: " + main_git + ' ' + test_cmd)
        __, __, exit_status = self._git_cmd(main_git, test_cmd, silent=True)
        if exit_status == 0:
            # sickrage.app.log.debug("Using: " + main_git)
            return main_git

        if platform.system().lower() in alternative_git:
            sickrage.app.log.debug("Trying known alternative GIT application locations")

            # sickrage.app.log.debug("Checking if we can use git commands: " + cur_git + ' ' + test_cmd)
            __, __, exit_status = self._git_cmd(alternative_git[platform.system().lower()], test_cmd)
            if exit_status == 0:
                # sickrage.app.log.debug("Using: " + cur_git)
                return alternative_git[platform.system().lower()]

        # Still haven't found a working git
        error_message = _('Unable to find your git executable - Set your git path from Settings->General->Advanced OR '
                          'delete your {git_folder} folder and run from source to enable '
                          'updates.'.format(**{'git_folder': os.path.join(sickrage.MAIN_DIR, '.git')}))

        sickrage.app.alerts.error(_('Updater'), error_message)

    @staticmethod
    def _git_cmd(git_path, args, silent=False):
        output = err = None

        if not git_path:
            sickrage.app.log.warning("No path to git specified, can't use git commands")
            exit_status = 1
            return output, err, exit_status

        cmd = [git_path] + args.split()

        try:
            if not silent:
                sickrage.app.log.debug("Executing " + ' '.join(cmd) + " with your shell in " + sickrage.MAIN_DIR)

            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                 shell=(sys.platform == 'win32'), cwd=sickrage.MAIN_DIR)
            output, err = p.communicate()
            exit_status = p.returncode

            if output is not None:
                output = output.decode("utf-8", "ignore").strip() if isinstance(output, bytes) else output.strip()
        except (RuntimeError, OSError):
            sickrage.app.log.info(f"Command {' '.join(cmd)} didn\'t work")
            exit_status = 1

        if exit_status == 0:
            if not silent:
                sickrage.app.log.debug(f"{' '.join(cmd)} : returned successful")
            exit_status = 0
        elif exit_status == 1:
            if output:
                if 'stash' in output:
                    sickrage.app.log.warning("Please enable 'git reset' in settings or stash your changes in local files")
                else:
                    sickrage.app.log.debug(f"{' '.join(cmd)} returned : {str(output)}")
            else:
                sickrage.app.log.warning(f'{cmd} returned no data')
            exit_status = 1
        elif exit_status == 128 or 'fatal:' in output or err:
            sickrage.app.log.debug(f"{' '.join(cmd)} returned : {str(output)}")
            exit_status = 128
        else:
            sickrage.app.log.debug(f"{' '.join(cmd)} returned : {str(output)}, treat as error for now")
            exit_status = 1

        return output, err, exit_status

    def need_update(self):
        try:
            return self.version != self.latest_version and self._num_commits_behind > 0
        except Exception as e:
            sickrage.app.log.error(f"Unable to contact server, can't check for update: {e!r}")
            return False

    def update(self):
        """
        Calls git pull origin <branch> in order to update SiCKRAGE. Returns a bool depending
        on the call's success.
        """

        if sickrage.app.config.general.git_reset:
            self.reset()

        if not self.upgrade_pip():
            return False

        if not self.install_requirements(self.current_branch):
            return False

        __, __, exit_status = self._git_cmd(self._git_path, f'pull -f {sickrage.app.git_remote_url} {self.current_branch}')
        if exit_status == 0:
            sickrage.app.log.info("Updating SiCKRAGE from GIT servers")
            sickrage.app.alerts.message(_('Updater'), _('Updating SiCKRAGE from GIT servers'))
            NotificationProvider.mass_notify_version_update(self.latest_version)
            return True

        return False

    def clean(self):
        """
        Calls git clean to remove all untracked files. Returns a bool depending
        on the call's success.
        """
        __, __, exit_status = self._git_cmd(self._git_path, 'clean -df ""')
        return (False, True)[exit_status == 0]

    def reset(self):
        """
        Calls git reset --hard to perform a hard reset. Returns a bool depending
        on the call's success.
        """
        __, __, exit_status = self._git_cmd(self._git_path, 'reset --hard')
        return (False, True)[exit_status == 0]

    def fetch(self):
        """
        Calls git fetch to fetch all remote branches
        on the call's success.
        """
        __, __, exit_status = self._git_cmd(self._git_path, f'config remote.origin.fetch {"+refs/heads/*:refs/remotes/origin/*"}')
        if exit_status == 0:
            __, __, exit_status = self._git_cmd(self._git_path, 'fetch --all')
        return (False, True)[exit_status == 0]

    def checkout_branch(self, branch):
        if branch in self.remote_branches:
            sickrage.app.log.debug(f"Branch checkout: {self.version} -> {branch}")

            if not self.upgrade_pip():
                return False

            if not self.install_requirements(self.current_branch):
                return False

            # remove untracked files and performs a hard reset on git branch to avoid update issues
            if sickrage.app.config.general.git_reset:
                self.reset()

            # fetch all branches
            self.fetch()

            __, __, exit_status = self._git_cmd(self._git_path, f'checkout -f {branch}')
            if exit_status == 0:
                return True

        return False

    def get_remote_url(self):
        url, __, exit_status = self._git_cmd(self._git_path, f'remote get-url {sickrage.app.git_remote_url}')
        return ("", url)[exit_status == 0 and url is not None]

    def set_remote_url(self):
        if not sickrage.app.developer:
            self._git_cmd(self._git_path, f'remote set-url {sickrage.app.git_remote_url} {sickrage.app.app.git_remote_url}')


class WindowsUpdateManager(UpdateManager):
    def __init__(self):
        super(WindowsUpdateManager, self).__init__()
        self.type = "windows"
        self.manual_update = True


class SynologyUpdateManager(UpdateManager):
    def __init__(self):
        super(SynologyUpdateManager, self).__init__()
        self.type = "synology"
        self.manual_update = True


class DockerUpdateManager(UpdateManager):
    def __init__(self):
        super(DockerUpdateManager, self).__init__()
        self.type = "docker"
        self.manual_update = True


class ReadynasUpdateManager(UpdateManager):
    def __init__(self):
        super(ReadynasUpdateManager, self).__init__()
        self.type = "readynas"
        self.manual_update = True


class QnapUpdateManager(UpdateManager):
    def __init__(self):
        super(QnapUpdateManager, self).__init__()
        self.type = "qnap"
        self.manual_update = True


class PipUpdateManager(UpdateManager):
    def __init__(self):
        super(PipUpdateManager, self).__init__()
        self.type = "pip"
        self.manual_update = True


class SourceUpdateManager(UpdateManager):
    def __init__(self):
        super(SourceUpdateManager, self).__init__()
        self.type = "source"

    def update(self):
        """
        Downloads the latest source tarball from server and installs it over the existing version.
        """

        latest_version = self.latest_version

        tar_download_url = f'https://git.sickrage.ca/SiCKRAGE/sickrage/-/archive/{latest_version}/sickrage-{latest_version}.tar.gz'

        try:
            if not self.upgrade_pip():
                return False

            if not self.install_requirements(self.current_branch):
                return False

            retry_count = 0
            while retry_count < 3:
                with tempfile.TemporaryFile() as update_tarfile:
                    sickrage.app.log.info(f"Downloading update from {tar_download_url!r}")
                    resp = WebSession().get(tar_download_url)
                    if not resp or not resp.content:
                        sickrage.app.log.warning('Failed to download SiCKRAGE update')
                        retry_count += 1
                        continue

                    update_tarfile.write(resp.content)
                    update_tarfile.seek(0)

                    with tempfile.TemporaryDirectory(prefix='sr_update_', dir=sickrage.app.data_dir) as unpack_dir:
                        sickrage.app.log.info("Extracting SiCKRAGE update file")
                        try:
                            tar = tarfile.open(fileobj=update_tarfile, mode='r:gz')
                            tar.extractall(unpack_dir)
                            tar.close()
                        except tarfile.TarError:
                            sickrage.app.log.warning("Invalid update data, update failed: not a gzip file")
                            retry_count += 1
                            continue

                        if len(os.listdir(unpack_dir)) != 1:
                            sickrage.app.log.warning("Invalid update data, update failed")
                            retry_count += 1
                            continue

                        update_dir = os.path.join(*[unpack_dir, os.listdir(unpack_dir)[0], 'sickrage'])
                        sickrage.app.log.info(f"Sync folder {update_dir} to {sickrage.PROG_DIR}")
                        dirsync.sync(update_dir, sickrage.PROG_DIR, 'sync', purge=True)

                        # Notify update successful
                        NotificationProvider.mass_notify_version_update(latest_version)

                        return True
        except Exception as e:
            sickrage.app.log.error(f"Error while trying to update: {e!r}")
            return False



"""Initial migration

Revision ID: 12
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '12'
down_revision = '11'


class SearchFormats(object):
    STANDARD = 1
    AIR_BY_DATE = 2
    ANIME = 3
    SPORTS = 4
    COLLECTION = 6

    search_format_strings = {
        STANDARD: 'Standard (Show.S01E01)',
        AIR_BY_DATE: 'Air By Date (Show.2010.03.02)',
        ANIME: 'Anime (Show.265)',
        SPORTS: 'Sports (Show.2010.03.02)',
        COLLECTION: 'Collection (Show.Series.1.1of10) or (Show.Series.1.Part.1)'
    }


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    tv_shows = sa.Table('tv_shows', meta, autoload=True)

    if not hasattr(tv_shows.c, 'search_format'):
        op.add_column('tv_shows', sa.Column('search_format', sa.Integer, default=0))

    with op.get_context().begin_transaction():
        for row in conn.execute(tv_shows.select()):
            if row.anime == 1 and not row.scene == 1:
                value = SearchFormats.ANIME
            elif row.anime == 1 and row.scene == 1:
                value = 5
            elif row.sports == 1:
                value = SearchFormats.SPORTS
            elif row.air_by_date == 1:
                value = SearchFormats.AIR_BY_DATE
            elif row.scene == 1:
                value = 5
            else:
                value = SearchFormats.STANDARD

            conn.execute(f'UPDATE tv_shows SET search_format = {value} WHERE tv_shows.indexer_id = {row.indexer_id}')

    with op.batch_alter_table('tv_shows') as batch_op:
        if hasattr(tv_shows.c, 'sports'):
            batch_op.drop_column('sports')

        if hasattr(tv_shows.c, 'air_by_date'):
            batch_op.drop_column('air_by_date')


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import os
import pathlib
import re
import subprocess

import subliminal
from babelfish import language_converters, Language
from subliminal import save_subtitles

import sickrage
from sickrage.core.helpers import chmod_as_parent, make_dir
from sickrage.core.tv.show.helpers import find_show
from sickrage.core.tv.show.history import History
from sickrage.subtitles.providers.utils import hash_itasa


class Subtitles(object):
    def __init__(self):
        # add subtitle providers
        for provider in ('itasa = sickrage.subtitles.providers.itasa:ItaSAProvider',
                         'legendastv = subliminal.providers.legendastv:LegendasTVProvider',
                         'wizdom = sickrage.subtitles.providers.wizdom:WizdomProvider',
                         'subscene = sickrage.subtitles.providers.subscene:SubsceneProvider',
                         'napiprojekt = subliminal.providers.napiprojekt:NapiProjektProvider'):
            if provider not in subliminal.provider_manager.registered_extensions + subliminal.provider_manager.internal_extensions:
                subliminal.provider_manager.register(str(provider))

        # add subtitle refiners
        for refiner in ('release = sickrage.subtitles.refiners.release:refine',
                        'tvepisode = sickrage.subtitles.refiners.tv_episode:refine'):
            if refiner not in subliminal.refiner_manager.registered_extensions + subliminal.refiner_manager.internal_extensions:
                subliminal.refiner_manager.register(str(refiner))

        # configure subtitle cache
        subliminal.region.configure('dogpile.cache.memory', replace_existing_backend=True)

        self.PROVIDER_URLS = {
            'addic7ed': 'http://www.addic7ed.com',
            'itasa': 'http://www.italiansubs.net/',
            'legendastv': 'http://www.legendas.tv',
            'napiprojekt': 'http://www.napiprojekt.pl',
            'opensubtitles': 'http://www.opensubtitles.org',
            'podnapisi': 'http://www.podnapisi.net',
            'subscenter': 'http://www.subscenter.org',
            'thesubdb': 'http://www.thesubdb.com',
            'tvsubtitles': 'http://www.tvsubtitles.net',
            'wizdom': 'http://wizdom.xyz',
            'subscene': 'https://subscene.com'
        }

        self.subtitle_extensions = [
            'srt',
            'sub',
            'ass',
            'idx',
            'ssa'
        ]

        self.episode_refiners = (
            'metadata',
            'release',
            'tvepisode',
            'tvdb',
            'omdb'
        )

    def sortedServiceList(self):
        newList = []
        lmgtfy = 'http://lmgtfy.com/?q=%s'

        curIndex = 0
        for curService in sickrage.app.config.subtitles.services_list.split(','):
            if curService in subliminal.provider_manager.names():
                newList.append({'name': curService,
                                'url': self.PROVIDER_URLS[curService] if curService in self.PROVIDER_URLS else lmgtfy % curService,
                                'image': curService + '.png',
                                'enabled': [int(x) for x in sickrage.app.config.subtitles.services_enabled.split('|') if x][curIndex] == 1
                                })
            curIndex += 1

        for curService in subliminal.provider_manager.names():
            if curService not in [x['name'] for x in newList]:
                newList.append({'name': curService,
                                'url': self.PROVIDER_URLS[curService] if curService in self.PROVIDER_URLS else lmgtfy % curService,
                                'image': curService + '.png',
                                'enabled': False,
                                })

        return newList

    def getEnabledServiceList(self):
        return [x['name'] for x in self.sortedServiceList() if x['enabled']]

    def download_subtitles(self, series_id, series_provider_id, season, episode):
        show_object = find_show(series_id, series_provider_id)
        if not show_object:
            return [], None

        episode_object = show_object.get_episode(season, episode)

        existing_subtitles = episode_object.subtitles
        if not isinstance(existing_subtitles, list):
            existing_subtitles = []

        # First of all, check if we need subtitles
        languages = self.get_needed_languages(existing_subtitles)
        if not languages:
            sickrage.app.log.debug('%s: No missing subtitles for S%02dE%02d' % (series_id, season, episode))
            return existing_subtitles, None

        subtitles_path = self.get_subtitles_path(episode_object.location)
        video_path = episode_object.location
        providers = self.getEnabledServiceList()

        video = self.get_video(video_path, subtitles_path=subtitles_path, episode_object=episode_object)
        if not video:
            sickrage.app.log.debug('%s: Exception caught in subliminal.scan_video for S%02dE%02d' % (series_id, season, episode))
            return existing_subtitles, None

        provider_configs = {
            'addic7ed': {
                'username': sickrage.app.config.subtitles.addic7ed_user,
                'password': sickrage.app.config.subtitles.addic7ed_pass
            },
            'itasa': {
                'username': sickrage.app.config.subtitles.itasa_user,
                'password': sickrage.app.config.subtitles.itasa_pass
            },
            'legendastv': {
                'username': sickrage.app.config.subtitles.legendastv_user,
                'password': sickrage.app.config.subtitles.legendastv_pass
            },
            'opensubtitles': {
                'username': sickrage.app.config.subtitles.opensubtitles_user,
                'password': sickrage.app.config.subtitles.opensubtitles_pass
            }
        }

        pool = subliminal.ProviderPool(providers=providers, provider_configs=provider_configs)

        try:
            subtitles_list = pool.list_subtitles(video, languages)
            if not subtitles_list:
                sickrage.app.log.debug('%s: No subtitles found for S%02dE%02d on any provider' % (series_id, season, episode))
                return existing_subtitles, None

            found_subtitles = pool.download_best_subtitles(subtitles_list, video, languages=languages,
                                                           hearing_impaired=sickrage.app.config.subtitles.hearing_impaired,
                                                           only_one=not sickrage.app.config.subtitles.multi)

            save_subtitles(video, found_subtitles, directory=subtitles_path, single=not sickrage.app.config.subtitles.multi)

            if not sickrage.app.config.subtitles.enable_embedded and sickrage.app.config.subtitles.extra_scripts and video_path.endswith(('.mkv', '.mp4')):
                self.run_subs_extra_scripts(episode_object, found_subtitles, video, single=not sickrage.app.config.subtitles.multi)

            new_subtitles = sorted({subtitle.language.opensubtitles for subtitle in found_subtitles})
            current_subtitles = sorted({subtitle for subtitle in new_subtitles + existing_subtitles if subtitle})
            if not sickrage.app.config.subtitles.multi and len(found_subtitles) == 1:
                new_code = found_subtitles[0].language.opensubtitles
                if new_code not in existing_subtitles:
                    current_subtitles.remove(new_code)
                current_subtitles.append('und')

        except Exception as e:
            sickrage.app.log.error("Error occurred when downloading subtitles for {}: {}".format(video_path, e))
            return existing_subtitles, None

        if sickrage.app.config.subtitles.history:
            for subtitle in found_subtitles:
                sickrage.app.log.debug('history.logSubtitle %s, %s' % (subtitle.provider_name, subtitle.language.opensubtitles))
                History.log_subtitle(series_id, series_provider_id, season, episode, episode_object.status, subtitle)

        return current_subtitles, new_subtitles

    def wanted_languages(self):
        return frozenset(sickrage.app.config.subtitles.languages.split(',')).intersection(self.subtitle_code_filter())

    def get_needed_languages(self, subtitles):
        if not sickrage.app.config.subtitles.multi:
            return set() if 'und' in subtitles else {self.from_code(language) for language in self.wanted_languages()}
        return {self.from_code(language) for language in self.wanted_languages().difference(subtitles)}

    def refresh_subtitles(self, series_id, series_provider_id, season, episode):
        show_object = find_show(series_id, series_provider_id)
        if not show_object:
            return [], None

        episode_object = show_object.get_episode(season, episode)

        video = self.get_video(episode_object.location, episode_object=episode_object)
        if not video:
            sickrage.app.log.debug("Exception caught in subliminal.scan_video, subtitles couldn't be refreshed")
            return episode_object.subtitles, None
        current_subtitles = self.get_subtitles(video)
        if episode_object.subtitles == ','.join(current_subtitles):
            sickrage.app.log.debug('No changed subtitles for {}'.format(episode_object.pretty_name()))
            return episode_object.subtitles, None
        else:
            return current_subtitles, True

    def get_video(self, video_path, subtitles_path=None, subtitles=True, embedded_subtitles=None, episode_object=None):
        if not subtitles_path:
            subtitles_path = self.get_subtitles_path(video_path)

        try:
            video = subliminal.scan_video(video_path)
        except Exception as error:
            sickrage.app.log.debug('Exception: {}'.format(error))
        else:
            if video.size > 10485760:
                video.hashes['itasa'] = hash_itasa(video_path)

            # external subtitles
            if subtitles:
                video.subtitle_languages |= set(subliminal.core.search_external_subtitles(video_path, directory=subtitles_path).values())

            if embedded_subtitles is None:
                embedded_subtitles = bool(
                    not sickrage.app.config.subtitles.enable_embedded and video_path.endswith('.mkv'))

            subliminal.refine(video,
                              episode_refiners=self.episode_refiners,
                              embedded_subtitles=embedded_subtitles,
                              release_name=episode_object.name,
                              tv_episode=episode_object)

            video.alternative_series = [x.split('|')[0] for x in episode_object.show.scene_exceptions]

            # remove format metadata
            video.format = ""

            return video

    def get_subtitles_path(self, video_path):
        if pathlib.Path(sickrage.app.config.subtitles.dir).is_absolute() and pathlib.Path(sickrage.app.config.subtitles.dir).exists():
            new_subtitles_path = sickrage.app.config.subtitles.dir
        elif sickrage.app.config.subtitles.dir:
            new_subtitles_path = str(pathlib.Path(os.path.dirname(video_path)).joinpath(sickrage.app.config.subtitles.dir.strip('/')))
            dir_exists = make_dir(new_subtitles_path)
            if not dir_exists:
                sickrage.app.log.warning('Unable to create subtitles folder {}'.format(new_subtitles_path))
            else:
                chmod_as_parent(new_subtitles_path)
        else:
            new_subtitles_path = str(pathlib.Path(video_path).parent)

        return new_subtitles_path

    def get_subtitles(self, video):
        """Return a sorted list of detected subtitles for the given video file."""
        result_list = []

        if not video.subtitle_languages:
            return result_list

        for language in video.subtitle_languages:
            if hasattr(language, 'opensubtitles') and language.opensubtitles:
                result_list.append(language.opensubtitles)

        return sorted(result_list)

    def scan_subtitle_languages(self, path):
        language_extensions = tuple('.' + c for c in language_converters['opensubtitles'].codes)
        dirpath, filename = os.path.split(path)
        subtitles = set()
        for p in os.listdir(dirpath):
            if not isinstance(p, bytes) and p.startswith(os.path.splitext(filename)[0]) and p.endswith(
                    subliminal.SUBTITLE_EXTENSIONS):
                if os.path.splitext(p)[0].endswith(language_extensions) and len(
                        os.path.splitext(p)[0].rsplit('.', 1)[1]) == 2:
                    subtitles.add(Language.fromopensubtitles(os.path.splitext(p)[0][-2:]))
                elif os.path.splitext(p)[0].endswith(language_extensions) and len(
                        os.path.splitext(p)[0].rsplit('.', 1)[1]) == 3:
                    subtitles.add(Language.fromopensubtitles(os.path.splitext(p)[0][-3:]))
                elif os.path.splitext(p)[0].endswith('pt-BR') and len(
                        os.path.splitext(p)[0].rsplit('.', 1)[1]) == 5:
                    subtitles.add(Language.fromopensubtitles('pob'))
                else:
                    subtitles.add(Language('und'))

        return subtitles

    def subtitle_code_filter(self):
        return {code for code in language_converters['opensubtitles'].codes if len(code) == 3}

    def from_code(self, language):
        language = language.strip()
        if language and language in language_converters['opensubtitles'].codes:
            return Language.fromopensubtitles(language)

        return Language('und')

    def name_from_code(self, code):
        return self.from_code(code).name

    def code_from_code(self, code):
        return self.from_code(code).opensubtitles

    def run_subs_extra_scripts(self, episode_object, found_subtitles, video, single=False):
        for curScriptName in sickrage.app.config.subtitles.extra_scripts.split('|'):
            script_cmd = [piece for piece in re.split("( |\\\".*?\\\"|'.*?')", curScriptName) if piece.strip()]
            script_cmd[0] = os.path.abspath(script_cmd[0])
            sickrage.app.log.debug("Absolute path to script: " + script_cmd[0])

            for subtitle in found_subtitles:
                subtitle_path = subliminal.subtitle.get_subtitle_path(video.name, None if single else subtitle.language)

                inner_cmd = script_cmd + [video.name, subtitle_path, subtitle.language.opensubtitles, episode_object.show.name,
                                          str(episode_object.season), str(episode_object.episode), episode_object.name, str(episode_object.show.series_id)]

                # use subprocess to run the command and capture output
                sickrage.app.log.info("Executing command: %s" % inner_cmd)
                try:
                    p = subprocess.Popen(inner_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                         stderr=subprocess.STDOUT, cwd=sickrage.PROG_DIR)
                    out, __ = p.communicate()
                    sickrage.app.log.debug("Script result: %s" % out)

                except Exception as e:
                    sickrage.app.log.info("Unable to run subs_extra_script: {}".format(e))
